from six.moves import range

import numpy as np
import scipy.sparse as sparse
import scipy.sparse.linalg as linalg
from landlab.grid.base import BAD_INDEX_VALUE

# these ones only so we can run this module ad-hoc:
# import pylab
from landlab import ModelParameterDictionary, Component
from landlab.utils.decorators import use_file_name_or_kwds
# from copy import copy

# Things to add: 1. Explicit stability check.
# 2. Implicit handling of scenarios where kappa*dt exceeds critical step -
#    subdivide dt automatically.


class PerronNLDiffuse(Component):
    '''
    This component implements nonlinear diffusion, following Perron (2011).

    This module uses Taylor Perron's implicit (2011) method to solve the
    nonlinear hillslope diffusion equation across a rectangular, regular grid
    for a single timestep. Note it works with the mass flux implicitly, and
    thus does not actually calculate it. Grid must be at least 5x5.

    Boundary condition handling assumes each edge uses the same BC for each of
    its nodes.
    This component cannot yet handle looped boundary conditions, but all others
    should be fine.

    This component has KNOWN STABILITY ISSUES which will be resolved in a
    future release; use at your own risk.

    The primary method of this class is :func:`run_one_step`.

    Construction::

        PerronNLDiffuse(grid, linear_diffusivity=None, S_crit=33.*np.pi/180.,
                        rock_density=2700., sed_density=2700.)

    Parameters
    ----------
    grid : RasterModelGrid
        A Landlab raster grid
    nonlinear_diffusivity : float, array or field name
        The nonlinear diffusivity
    S_crit : float (radians)
        The critical hillslope angle
    rock_density : float (kg*m**-3)
        The density of intact rock
    sed_density : float (kg*m**-3)
        The density of the mobile (sediment) layer

    Examples
    --------
    >>> from landlab.components import PerronNLDiffuse
    >>> from landlab import RasterModelGrid
    >>> import numpy as np
    >>> mg = RasterModelGrid((5, 5))
    >>> z = mg.add_zeros('node', 'topographic__elevation')
    >>> nl = PerronNLDiffuse(mg, nonlinear_diffusivity=1.)
    >>> dt = 100.
    >>> nt = 20
    >>> uplift_rate = 0.001
    >>> for i in range(nt):
    ...     z[mg.core_nodes] += uplift_rate*dt
    ...     nl.run_one_step(dt)
    >>> z_target = np.array(
    ...     [ 0.        ,  0.        ,  0.        ,  0.        ,  0.        ,
    ...       0.        ,  0.00778637,  0.0075553 ,  0.00778637,  0.        ,
    ...       0.        ,  0.0075553 ,  0.0078053 ,  0.0075553 ,  0.        ,
    ...       0.        ,  0.00778637,  0.0075553 ,  0.00778637,  0.        ,
    ...       0.        ,  0.        ,  0.        ,  0.        ,  0.        ])
    >>> np.allclose(z, z_target)
    True
    '''

    _name = 'PerronNLDiffuse'

    _input_var_names = ('topographic__elevation', )

    _output_var_names = ('topographic__elevation', )

    _var_units = {'topographic__elevation': 'm'}

    _var_mapping = {'topographic__elevation': 'node'}

    _var_doc = {
        'topographic__elevation': ('Land surface topographic elevation; can ' +
                                   'be overwritten in initialization')}

    @use_file_name_or_kwds
    def __init__(self, grid, nonlinear_diffusivity=None, S_crit=33.*np.pi/180.,
                 rock_density=2700., sed_density=2700., **kwds):
        # disable internal_uplift option:
        internal_uplift = None
        self._grid = grid
        self.values_to_diffuse = 'topographic__elevation'
        if nonlinear_diffusivity is not None:
            if nonlinear_diffusivity is not str:
                self._kappa = nonlinear_diffusivity
            else:
                self._kappa = self.grid.at_node[nonlinear_diffusivity]
        else:
            try:
                self._kappa = kwds.pop('kappa', None)
            except KeyError:
                raise KeyError("nonlinear_diffusivity must be provided to " +
                               "the PerronNLDiffuse component")
        if internal_uplift is None:
            self.internal_uplifts = False
            self._uplift = 0.
        else:
            self.internal_uplifts = True
            self._uplift = float(internal_uplift)
            # self._uplift = self.grid.zeros('node', dtype=float)
            # self._uplift[self.grid.core_nodes] = internal_uplift
        self._rock_density = rock_density
        self._sed_density = sed_density
        self._S_crit = S_crit

        # for component back compatibility (undocumented):
        # ###
        self.timestep_in = kwds.pop('dt', None)
        if 'values_to_diffuse' in kwds.keys():
            self.values_to_diffuse = kwds.pop('values_to_diffuse')
            for mytups in (self._input_var_names, self._output_var_names):
                myset = set(mytups)
                myset.remove('topographic__elevation')
                myset.add(self.values_to_diffuse)
                mytups = tuple(myset)
            for mydicts in (self._var_units, self._var_mapping, self._var_doc):
                mydicts[self.values_to_diffuse] = mydicts.pop(
                    'topographic__elevation')

        self._delta_x = grid.dx
        self._delta_y = grid.dy
        self._one_over_delta_x = 1. / self._delta_x
        self._one_over_delta_y = 1. / self._delta_y
        self._one_over_delta_x_sqd = self._one_over_delta_x**2.
        self._one_over_delta_y_sqd = self._one_over_delta_y**2.
        self._b = 1. / self._S_crit**2.

        ncols = grid.number_of_node_columns
        self.ncols = ncols
        nrows = grid.number_of_node_rows
        self.nrows = nrows
        nnodes = grid.number_of_nodes
        self.nnodes = nnodes
        ninteriornodes = grid.number_of_interior_nodes
        ncorenodes = ninteriornodes - 2 * (ncols + nrows - 6)
        self.ninteriornodes = ninteriornodes
        self.interior_grid_width = ncols - 2
        self.core_cell_width = ncols - 4

        self._interior_corners = np.array([ncols + 1, 2 * ncols - 2,
                                           nnodes - 2 * ncols + 1,
                                           nnodes - ncols - 2])
        _left_list = np.array(
            range(2 * ncols + 1, nnodes - 2 * ncols, ncols))
        # ^these are still real IDs
        _right_list = np.array(
            range(3 * ncols - 2, nnodes - 2 * ncols, ncols))
        _bottom_list = np.array(range(ncols + 2, 2 * ncols - 2))
        _top_list = np.array(
            range(nnodes - 2 * ncols + 2, nnodes - ncols - 2))
        self._left_list = _left_list
        self._right_list = _right_list
        self._bottom_list = _bottom_list
        self._top_list = _top_list

        self._core_nodes = self._coreIDtoreal(np.arange(
            ncorenodes, dtype=int))
        self.corenodesbyintIDs = self._realIDtointerior(self._core_nodes)
        self.ncorenodes = len(self._core_nodes)

        self.corner_interior_IDs = self._realIDtointerior(
            self._interior_corners)
        # ^i.e., interior corners as interior IDs
        self.bottom_interior_IDs = self._realIDtointerior(np.array(
            _bottom_list))
        self.top_interior_IDs = self._realIDtointerior(np.array(_top_list))
        self.left_interior_IDs = self._realIDtointerior(np.array(_left_list))
        self.right_interior_IDs = self._realIDtointerior(np.array(
            _right_list))

        # build an ID map to let us easily map the variables of the core nodes
        # onto the operating matrix:
        # This array is ninteriornodes long, but the IDs it contains are
        # REAL IDs
        operating_matrix_ID_map = np.empty((ninteriornodes, 9))
        self.interior_IDs_as_real = self._interiorIDtoreal(
            np.arange(ninteriornodes))
        for j in range(ninteriornodes):
            i = self.interior_IDs_as_real[j]
            operating_matrix_ID_map[j, :] = np.array(
                [(i-ncols-1), (i-ncols), (i-ncols+1), (i-1), i, (i+1),
                 (i+ncols-1), (i+ncols), (i+ncols+1)])
        self.operating_matrix_ID_map = operating_matrix_ID_map
        self.operating_matrix_core_int_IDs = self._realIDtointerior(
            operating_matrix_ID_map[self.corenodesbyintIDs, :])
        # ^shape(ncorenodes,9)
        # see below for corner and edge maps

        # Build masks for the edges and corners to be applied to the operating
        # matrix map.
        # Antimasks are the boundary nodes, masks are "normal"
        topleft_mask = [1, 2, 4, 5]
        topleft_antimask = [0, 3, 6, 7, 8]
        topright_mask = [0, 1, 3, 4]
        topright_antimask = [2, 5, 6, 7, 8]
        bottomleft_mask = [4, 5, 7, 8]
        bottomleft_antimask = [0, 1, 2, 3, 6]
        bottomright_mask = [3, 4, 6, 7]
        bottomright_antimask = [0, 1, 2, 5, 8]
        self.corners_masks = (np.vstack((bottomleft_mask, bottomright_mask,
                                         topleft_mask, topright_mask)))
        # ^(each_corner,mask_for_each_corner)
        self.corners_antimasks = (np.vstack((bottomleft_antimask,
                                             bottomright_antimask,
                                             topleft_antimask,
                                             topright_antimask)))
        # ^so shape becomes (4,5)
        self.left_mask = [1, 2, 4, 5, 7, 8]
        self.left_antimask = [0, 3, 6]
        self.top_mask = [0, 1, 2, 3, 4, 5]
        self.top_antimask = [6, 7, 8]
        self.right_mask = [0, 1, 3, 4, 6, 7]
        self.right_antimask = [2, 5, 8]
        self.bottom_mask = [3, 4, 5, 6, 7, 8]
        self.bottom_antimask = [0, 1, 2]
        self.antimask_corner_position = [0, 2, 2, 4]
        # ^this is the position w/i the corner antimasks that the true corner
        # actually occupies

        self.modulator_mask = np.array([-ncols - 1, -ncols, -ncols + 1, -1,
                                        0, 1, ncols - 1, ncols, ncols + 1])

        # ^Set up terms for BC handling (still feels very clumsy)
        bottom_edge = grid.nodes_at_bottom_edge[1: -1]
        top_edge = grid.nodes_at_top_edge[1: -1]
        left_edge = grid.nodes_at_left_edge[1: -1]
        right_edge = grid.nodes_at_right_edge[1: -1]
        self.bottom_flag = 1
        self.top_flag = 1
        self.left_flag = 1
        self.right_flag = 1
        # self.corner_flags = [1,1,1,1] #In ID order, so BL,BR,TL,TR
        if np.all(grid.status_at_node[bottom_edge] == 4):
            # ^This should be all of them, or none of them
            self.bottom_flag = 4
        elif np.all(grid.status_at_node[bottom_edge] == 3):
            self.bottom_flag = 3
        elif np.all(grid.status_at_node[bottom_edge] == 2):
            self.bottom_flag = 2
        elif np.all(grid.status_at_node[bottom_edge] == 1):
            pass
        else:
            raise NameError("Different cells on the same grid edge have "
                            "different boundary statuses")
            # Note this could get fraught if we need to open a cell to let
            # water flow out...
        if np.all(grid.status_at_node[top_edge] == 4):
            self.top_flag = 4
        elif np.all(grid.status_at_node[top_edge] == 3):
            self.top_flag = 3
        elif np.all(grid.status_at_node[top_edge] == 2):
            self.top_flag = 2
        elif np.all(grid.status_at_node[top_edge] == 1):
            pass
        else:
            raise NameError("Different cells on the same grid edge have "
                            "different boundary statuses")
        if np.all(grid.status_at_node[left_edge] == 4):
            self.left_flag = 4
        elif np.all(grid.status_at_node[left_edge] == 3):
            self.left_flag = 3
        elif np.all(grid.status_at_node[left_edge] == 2):
            self.left_flag = 2
        elif np.all(grid.status_at_node[left_edge] == 1):
            pass
        else:
            raise NameError("Different cells on the same grid edge have "
                            "different boundary statuses")
        if np.all(grid.status_at_node[right_edge] == 4):
            self.right_flag = 4
        elif np.all(grid.status_at_node[right_edge] == 3):
            self.right_flag = 3
        elif np.all(grid.status_at_node[right_edge] == 2):
            self.right_flag = 2
        elif np.all(grid.status_at_node[right_edge] == 1):
            pass
        else:
            raise NameError("Different cells on the same grid edge have "
                            "different boundary statuses")

        self.fixed_grad_BCs_present = (self.bottom_flag == 2 or
                                       self.top_flag == 2 or
                                       self.left_flag == 2 or
                                       self.right_flag == 2)
        self.looped_BCs_present = (self.bottom_flag == 3 or
                                   self.top_flag == 3 or
                                   self.left_flag == 3 or
                                   self.right_flag == 3)
        if self.fixed_grad_BCs_present:
            if self.values_to_diffuse != grid.fixed_gradient_of:
                raise ValueError("Boundary conditions set in the grid don't "
                                 "apply to the data the diffuser is trying to "
                                 "work with")

        if np.any(grid.status_at_node == 2):
            self.fixed_grad_offset_map = np.empty(
                nrows * ncols, dtype=float)
            self.fixed_grad_anchor_map = np.empty_like(
                self.fixed_grad_offset_map)
            self.fixed_grad_offset_map[grid.fixed_gradient_node_properties[
                'boundary_node_IDs']] = grid.fixed_gradient_node_properties[
                'values_to_add']

        self.corner_flags = grid.status_at_node[[0, ncols - 1, -ncols, -1]]

        op_mat_just_corners = operating_matrix_ID_map[
            self.corner_interior_IDs, :]
        op_mat_cnr0 = op_mat_just_corners[0, bottomleft_mask]
        op_mat_cnr1 = op_mat_just_corners[1, bottomright_mask]
        op_mat_cnr2 = op_mat_just_corners[2, topleft_mask]
        op_mat_cnr3 = op_mat_just_corners[3, topright_mask]
        op_mat_just_active_cnrs = np.vstack((op_mat_cnr0, op_mat_cnr1,
                                             op_mat_cnr2, op_mat_cnr3))
        self.operating_matrix_corner_int_IDs = self._realIDtointerior(
            op_mat_just_active_cnrs)
        # ^(4corners,4nodesactivepercorner)
        self.operating_matrix_bottom_int_IDs = self._realIDtointerior(
            operating_matrix_ID_map[
                self.bottom_interior_IDs, :][:, self.bottom_mask])
        # ^(nbottomnodes,6activenodeseach)
        self.operating_matrix_top_int_IDs = self._realIDtointerior(
            operating_matrix_ID_map[
                self.top_interior_IDs, :][:, self.top_mask])
        self.operating_matrix_left_int_IDs = self._realIDtointerior(
            operating_matrix_ID_map[
                self.left_interior_IDs, :][:, self.left_mask])
        self.operating_matrix_right_int_IDs = self._realIDtointerior(
            operating_matrix_ID_map[
                self.right_interior_IDs, :][:, self.right_mask])

    def _initialize(self, grid, input_stream):
        inputs = ModelParameterDictionary(input_stream)
        self.inputs = inputs
        self.grid = grid

        self.internal_uplifts = False

        if self.internal_uplifts:
            try:
                self._uplift = inputs.read_float('uplift')
            except:
                self._uplift = inputs.read_float('uplift_rate')
        else:
            self._uplift = 0.
        self._rock_density = inputs.read_float('rock_density')
        self._sed_density = inputs.read_float('sed_density')
        self._kappa = inputs.read_float('kappa')  # ==_a
        self._S_crit = inputs.read_float('S_crit')
        try:
            self.values_to_diffuse = inputs.read_str('values_to_diffuse')
        except:
            self.values_to_diffuse = 'topographic__elevation'
        try:
            self.timestep_in = inputs.read_float('dt')
        except:
            raise NameError('''No fixed timestep supplied, it must be set
                       dynamically somewhere else. Be sure to call
                       input_timestep(timestep_in) as part of your run
                       loop.''')

        self._delta_x = grid.dx
        self._delta_y = grid.dy
        self._one_over_delta_x = 1. / self._delta_x
        self._one_over_delta_y = 1. / self._delta_y
        self._one_over_delta_x_sqd = self._one_over_delta_x**2.
        self._one_over_delta_y_sqd = self._one_over_delta_y**2.
        self._b = 1. / self._S_crit**2.

        ncols = grid.number_of_node_columns
        self.ncols = ncols
        nrows = grid.number_of_node_rows
        self.nrows = nrows
        nnodes = grid.number_of_nodes
        self.nnodes = nnodes
        ninteriornodes = grid.number_of_interior_nodes
        ncorenodes = ninteriornodes - 2 * (ncols + nrows - 6)
        self.ninteriornodes = ninteriornodes
        self.interior_grid_width = ncols - 2
        self.core_cell_width = ncols - 4

        self._interior_corners = np.array([ncols + 1, 2 * ncols - 2,
                                           nnodes - 2 * ncols + 1,
                                           nnodes - ncols - 2])
        _left_list = np.array(
            range(2 * ncols + 1, nnodes - 2 * ncols, ncols))
        # ^these are still real IDs
        _right_list = np.array(
            range(3 * ncols - 2, nnodes - 2 * ncols, ncols))
        _bottom_list = np.array(range(ncols + 2, 2 * ncols - 2))
        _top_list = np.array(
            range(nnodes - 2 * ncols + 2, nnodes - ncols - 2))
        self._left_list = _left_list
        self._right_list = _right_list
        self._bottom_list = _bottom_list
        self._top_list = _top_list

        self._core_nodes = self._coreIDtoreal(np.arange(
            ncorenodes, dtype=int))
        self.corenodesbyintIDs = self._realIDtointerior(self._core_nodes)
        self.ncorenodes = len(self._core_nodes)

        self.corner_interior_IDs = self._realIDtointerior(
            self._interior_corners)
        # ^i.e., interior corners as interior IDs
        self.bottom_interior_IDs = self._realIDtointerior(np.array(
            _bottom_list))
        self.top_interior_IDs = self._realIDtointerior(np.array(_top_list))
        self.left_interior_IDs = self._realIDtointerior(np.array(_left_list))
        self.right_interior_IDs = self._realIDtointerior(np.array(
            _right_list))

        # build an ID map to let us easily map the variables of the core nodes
        # onto the operating matrix:
        # This array is ninteriornodes long, but the IDs it contains are
        # REAL IDs
        operating_matrix_ID_map = np.empty((ninteriornodes, 9))
        self.interior_IDs_as_real = self._interiorIDtoreal(
            np.arange(ninteriornodes))
        for j in range(ninteriornodes):
            i = self.interior_IDs_as_real[j]
            operating_matrix_ID_map[j, :] = np.array(
                [(i-ncols-1), (i-ncols), (i-ncols+1), (i-1), i, (i+1),
                 (i+ncols-1), (i+ncols), (i+ncols+1)])
        self.operating_matrix_ID_map = operating_matrix_ID_map
        self.operating_matrix_core_int_IDs = self._realIDtointerior(
            operating_matrix_ID_map[self.corenodesbyintIDs, :])
        # ^shape(ncorenodes,9)
        # see below for corner and edge maps

        # Build masks for the edges and corners to be applied to the operating
        # matrix map.
        # Antimasks are the boundary nodes, masks are "normal"
        topleft_mask = [1, 2, 4, 5]
        topleft_antimask = [0, 3, 6, 7, 8]
        topright_mask = [0, 1, 3, 4]
        topright_antimask = [2, 5, 6, 7, 8]
        bottomleft_mask = [4, 5, 7, 8]
        bottomleft_antimask = [0, 1, 2, 3, 6]
        bottomright_mask = [3, 4, 6, 7]
        bottomright_antimask = [0, 1, 2, 5, 8]
        self.corners_masks = (np.vstack((bottomleft_mask, bottomright_mask,
                                         topleft_mask, topright_mask)))
        # ^(each_corner,mask_for_each_corner)
        self.corners_antimasks = (np.vstack((bottomleft_antimask,
                                             bottomright_antimask,
                                             topleft_antimask,
                                             topright_antimask)))
        # ^so shape becomes (4,5)
        self.left_mask = [1, 2, 4, 5, 7, 8]
        self.left_antimask = [0, 3, 6]
        self.top_mask = [0, 1, 2, 3, 4, 5]
        self.top_antimask = [6, 7, 8]
        self.right_mask = [0, 1, 3, 4, 6, 7]
        self.right_antimask = [2, 5, 8]
        self.bottom_mask = [3, 4, 5, 6, 7, 8]
        self.bottom_antimask = [0, 1, 2]
        self.antimask_corner_position = [0, 2, 2, 4]
        # ^this is the position w/i the corner antimasks that the true corner
        # actually occupies

        self.modulator_mask = np.array([-ncols - 1, -ncols, -ncols + 1, -1,
                                        0, 1, ncols - 1, ncols, ncols + 1])

        # ^Set up terms for BC handling (still feels very clumsy)
        bottom_edge = grid.nodes_at_bottom_edge[1: -1]
        top_edge = grid.nodes_at_top_edge[1: -1]
        left_edge = grid.nodes_at_left_edge[1: -1]
        right_edge = grid.nodes_at_right_edge[1: -1]
        self.bottom_flag = 1
        self.top_flag = 1
        self.left_flag = 1
        self.right_flag = 1
        # self.corner_flags = [1,1,1,1] #In ID order, so BL,BR,TL,TR
        if np.all(grid.status_at_node[bottom_edge] == 4):
            # ^This should be all of them, or none of them
            self.bottom_flag = 4
        elif np.all(grid.status_at_node[bottom_edge] == 3):
            self.bottom_flag = 3
        elif np.all(grid.status_at_node[bottom_edge] == 2):
            self.bottom_flag = 2
        elif np.all(grid.status_at_node[bottom_edge] == 1):
            pass
        else:
            raise NameError("Different cells on the same grid edge have "
                            "different boundary statuses")
            # Note this could get fraught if we need to open a cell to let
            # water flow out...
        if np.all(grid.status_at_node[top_edge] == 4):
            self.top_flag = 4
        elif np.all(grid.status_at_node[top_edge] == 3):
            self.top_flag = 3
        elif np.all(grid.status_at_node[top_edge] == 2):
            self.top_flag = 2
        elif np.all(grid.status_at_node[top_edge] == 1):
            pass
        else:
            raise NameError("Different cells on the same grid edge have "
                            "different boundary statuses")
        if np.all(grid.status_at_node[left_edge] == 4):
            self.left_flag = 4
        elif np.all(grid.status_at_node[left_edge] == 3):
            self.left_flag = 3
        elif np.all(grid.status_at_node[left_edge] == 2):
            self.left_flag = 2
        elif np.all(grid.status_at_node[left_edge] == 1):
            pass
        else:
            raise NameError("Different cells on the same grid edge have "
                            "different boundary statuses")
        if np.all(grid.status_at_node[right_edge] == 4):
            self.right_flag = 4
        elif np.all(grid.status_at_node[right_edge] == 3):
            self.right_flag = 3
        elif np.all(grid.status_at_node[right_edge] == 2):
            self.right_flag = 2
        elif np.all(grid.status_at_node[right_edge] == 1):
            pass
        else:
            raise NameError("Different cells on the same grid edge have "
                            "different boundary statuses")

        self.fixed_grad_BCs_present = (self.bottom_flag == 2 or
                                       self.top_flag == 2 or
                                       self.left_flag == 2 or
                                       self.right_flag == 2)
        self.looped_BCs_present = (self.bottom_flag == 3 or
                                   self.top_flag == 3 or
                                   self.left_flag == 3 or
                                   self.right_flag == 3)
        if self.fixed_grad_BCs_present:
            if self.values_to_diffuse != grid.fixed_gradient_of:
                raise ValueError("Boundary conditions set in the grid don't "
                                 "apply to the data the diffuser is trying to "
                                 "work with")

        if np.any(grid.status_at_node == 2):
            self.fixed_grad_offset_map = np.empty(
                nrows * ncols, dtype=float)
            self.fixed_grad_anchor_map = np.empty_like(
                self.fixed_grad_offset_map)
            self.fixed_grad_offset_map[grid.fixed_gradient_node_properties[
                'boundary_node_IDs']] = grid.fixed_gradient_node_properties[
                'values_to_add']

        self.corner_flags = grid.status_at_node[[0, ncols - 1, -ncols, -1]]

        op_mat_just_corners = operating_matrix_ID_map[self.corner_interior_IDs,
                                                      :]
        op_mat_cnr0 = op_mat_just_corners[0, bottomleft_mask]
        op_mat_cnr1 = op_mat_just_corners[1, bottomright_mask]
        op_mat_cnr2 = op_mat_just_corners[2, topleft_mask]
        op_mat_cnr3 = op_mat_just_corners[3, topright_mask]
        op_mat_just_active_cnrs = np.vstack((op_mat_cnr0, op_mat_cnr1,
                                            op_mat_cnr2, op_mat_cnr3))
        self.operating_matrix_corner_int_IDs = self._realIDtointerior(
            op_mat_just_active_cnrs)
        # ^(4corners,4nodesactivepercorner)
        self.operating_matrix_bottom_int_IDs = self._realIDtointerior(
            operating_matrix_ID_map[
                self.bottom_interior_IDs, :][:, self.bottom_mask])
        # ^(nbottomnodes,6activenodeseach)
        self.operating_matrix_top_int_IDs = self._realIDtointerior(
            operating_matrix_ID_map[
                self.top_interior_IDs, :][:, self.top_mask])
        self.operating_matrix_left_int_IDs = self._realIDtointerior(
            operating_matrix_ID_map[
                self.left_interior_IDs, :][:, self.left_mask])
        self.operating_matrix_right_int_IDs = self._realIDtointerior(
            operating_matrix_ID_map[
                self.right_interior_IDs, :][:, self.right_mask])

    def input_timestep(self, timestep_in):
        """
        Allows the user to set a dynamic (evolving) timestep manually as part
        of a run loop.
        """
        self.timestep_in = timestep_in

    def _gear_timestep(self, timestep_in, new_grid):
        """
        This method allows the gearing between the model run step and the
        component (shorter) step.
        The method becomes unstable if S>Scrit, so we test to prevent this.
        We implicitly assume the initial condition does not contain
        slopes > Scrit. If the method persistently explodes, this may be the
        problem.
        """
        extended_elevs = np.empty(
            self.grid.number_of_nodes + 1, dtype=float)
        extended_elevs[-1] = np.nan
        node_neighbors = self.grid.active_neighbors_at_node(bad_index=-1)
        extended_elevs[:-1] = new_grid['node'][self.values_to_diffuse]
        max_offset = np.nanmax(np.fabs(
            extended_elevs[:-1][node_neighbors] -
            extended_elevs[:-1].reshape((self.grid.number_of_nodes, 1))))
        if max_offset > np.tan(self._S_crit) * min(self.grid.dx,
                                                   self.grid.dy):
            # ^using S not tan(S) adds a buffer - but not appropriate
            self.internal_repeats = int(max_offset // (
                np.tan(self._S_crit) * min(self.grid.dx, self.grid.dy))) + 1
            # now we rig it so the actual timestep is an integer divisor
            # of T_in:
            self._delta_t = timestep_in / self.internal_repeats
            self.uplift_per_step = (new_grid['node'][self.values_to_diffuse] -
                                    self.grid['node'][self.values_to_diffuse]
                                    ) / self.internal_repeats
            if self.internal_repeats > 10000:
                raise ValueError('''Uplift rate is too high; solution is not
                                 stable!!''')
        else:
            self.internal_repeats = 1
            self._delta_t = timestep_in
            self.uplift_per_step = new_grid['node'][
                self.values_to_diffuse] - self.grid['node'][
                self.values_to_diffuse]
        return self._delta_t

    def _set_variables(self, grid):
        '''
        This function sets the variables needed for update().
        Now vectorized, shouold run faster.
        At the moment, this method can only handle fixed value BCs.
        '''
        n_interior_nodes = grid.number_of_interior_nodes

        # Initialize the local builder lists
        _mat_RHS = np.zeros(n_interior_nodes)

        try:
            elev = grid['node'][self.values_to_diffuse]
        except:
            raise NameError('elevations not found in grid!')
        try:
            _delta_t = self._delta_t
        except:
            raise NameError('''Timestep not set! Call _gear_timestep(tstep)
                            after initializing the component, but before
                            running it.''')
        _one_over_delta_x = self._one_over_delta_x
        _one_over_delta_x_sqd = self._one_over_delta_x_sqd
        _one_over_delta_y = self._one_over_delta_y
        _one_over_delta_y_sqd = self._one_over_delta_y_sqd
        _kappa = self._kappa
        _b = self._b
        _S_crit = self._S_crit
        _core_nodes = self._core_nodes
        corenodesbyintIDs = self.corenodesbyintIDs
        operating_matrix_core_int_IDs = self.operating_matrix_core_int_IDs
        operating_matrix_corner_int_IDs = self.operating_matrix_corner_int_IDs
        _interior_corners = self._interior_corners
        corners_antimasks = self.corners_antimasks
        corner_interior_IDs = self.corner_interior_IDs
        modulator_mask = self.modulator_mask
        corner_flags = self.corner_flags
        bottom_interior_IDs = self.bottom_interior_IDs
        top_interior_IDs = self.top_interior_IDs
        left_interior_IDs = self.left_interior_IDs
        right_interior_IDs = self.right_interior_IDs
        bottom_antimask = self.bottom_antimask
        _bottom_list = self._bottom_list
        top_antimask = self.top_antimask
        _top_list = self._top_list
        left_antimask = self.left_antimask
        _left_list = self._left_list
        right_antimask = self.right_antimask
        _right_list = self._right_list

        # Need to modify the "effective" values of the edge nodes if any of
        # the edges are inactive:
        if self.bottom_flag == 4:
            bottom_edge, inside_bottom_edge = grid.nodes[(0, 1), :]
            elev[bottom_edge] = elev[inside_bottom_edge]
            # corners are special cases, and assumed linked to the bottom and
            # top edge BCs...
            elev[bottom_edge[0]] = elev[inside_bottom_edge[1]]
            elev[bottom_edge[-1]] = elev[inside_bottom_edge[-2]]
        if self.top_flag == 4:
            top_edge, inside_top_edge = grid.nodes[(-1, -2), :]
            elev[top_edge] = elev[inside_top_edge]
            # corners are special cases, and assumed linked to the bottom and
            # top edge BCs...
            elev[top_edge[0]] = elev[inside_top_edge[1]]
            elev[top_edge[-1]] = elev[inside_top_edge[-2]]
        if self.left_flag == 4:
            left_edge = grid.nodes[1: -1, 0]
            inside_left_edge = grid.nodes[1: -1, 1]
            elev[left_edge] = elev[inside_left_edge]
        if self.right_flag == 4:
            right_edge = grid.nodes[1: -1, -1]
            inside_right_edge = grid.nodes[1: -1, -2]
            elev[right_edge] = elev[inside_right_edge]

        # replacing loop:
        cell_neighbors = grid.active_neighbors_at_node(bad_index=-1)
        # ^E,N,W,S
        cell_diagonals = grid._get_diagonal_list()  # NE,NW,SW,SE
        cell_neighbors[cell_neighbors == BAD_INDEX_VALUE] = -1
        cell_diagonals[cell_diagonals == BAD_INDEX_VALUE] = -1

        _z_x = (elev[cell_neighbors[:, 0]] - elev[cell_neighbors[:, 2]]
                ) * 0.5 * _one_over_delta_x
        _z_y = (elev[cell_neighbors[:, 1]] - elev[cell_neighbors[:, 3]]
                ) * 0.5 * _one_over_delta_y
        _z_xx = (elev[cell_neighbors[:, 0]] - 2. * elev + elev[
                 cell_neighbors[:, 2]]) * _one_over_delta_x_sqd
        _z_yy = (elev[cell_neighbors[:, 1]] - 2. * elev + elev[
                 cell_neighbors[:, 3]]) * _one_over_delta_y_sqd
        _z_xy = (elev[cell_diagonals[:, 0]] - elev[cell_diagonals[:, 1]] -
                 elev[cell_diagonals[:, 3]] + elev[cell_diagonals[:, 2]]
                 ) * 0.25 * _one_over_delta_x * _one_over_delta_y
        _d = 1. / (1. - _b * (_z_x * _z_x + _z_y * _z_y))

        _abd_sqd = _kappa * _b * _d * _d
        _F_ij = (-2.*_kappa*_d*(_one_over_delta_x_sqd+_one_over_delta_y_sqd) -
                 4.*_abd_sqd*(_z_x*_z_x*_one_over_delta_x_sqd +
                              _z_y*_z_y*_one_over_delta_y_sqd))
        _F_ijminus1 = (
            _kappa*_d*_one_over_delta_x_sqd - _abd_sqd*_z_x*(_z_xx+_z_yy) *
            _one_over_delta_x - 4.*_abd_sqd*_b*_d*(_z_x*_z_x*_z_xx+_z_y*_z_y *
                                                   _z_yy+2.*_z_x*_z_y*_z_xy) *
            _z_x*_one_over_delta_x - 2.*_abd_sqd*(
                _z_x*_z_xx*_one_over_delta_x -
                _z_x*_z_x*_one_over_delta_x_sqd +
                _z_y*_z_xy*_one_over_delta_x))
        _F_ijplus1 = (
            _kappa*_d*_one_over_delta_x_sqd + _abd_sqd*_z_x*(_z_xx+_z_yy) *
            _one_over_delta_x + 4.*_abd_sqd*_b*_d*(_z_x*_z_x*_z_xx+_z_y*_z_y *
                                                   _z_yy+2.*_z_x*_z_y*_z_xy) *
            _z_x*_one_over_delta_x + 2.*_abd_sqd*(
                _z_x*_z_xx*_one_over_delta_x +
                _z_x*_z_x*_one_over_delta_x_sqd +
                _z_y*_z_xy*_one_over_delta_x))
        _F_iminus1j = (
            _kappa*_d*_one_over_delta_y_sqd - _abd_sqd*_z_y*(_z_xx+_z_yy) *
            _one_over_delta_y - 4.*_abd_sqd*_b*_d*(_z_x*_z_x*_z_xx+_z_y*_z_y *
                                                   _z_yy+2.*_z_x*_z_y*_z_xy) *
            _z_y*_one_over_delta_y - 2.*_abd_sqd*(
                _z_y*_z_yy*_one_over_delta_y -
                _z_y*_z_y*_one_over_delta_y_sqd +
                _z_x*_z_xy*_one_over_delta_y))
        _F_iplus1j = (
            _kappa*_d*_one_over_delta_y_sqd + _abd_sqd*_z_y*(_z_xx+_z_yy) *
            _one_over_delta_y + 4.*_abd_sqd*_b*_d*(_z_x*_z_x*_z_xx+_z_y*_z_y *
                                                   _z_yy+2.*_z_x*_z_y*_z_xy) *
            _z_y*_one_over_delta_y + 2.*_abd_sqd*(
                _z_y*_z_yy*_one_over_delta_y +
                _z_y*_z_y*_one_over_delta_y_sqd +
                _z_x*_z_xy*_one_over_delta_y))
        _F_iplus1jplus1 = (
            _abd_sqd*_z_x*_z_y*_one_over_delta_x*_one_over_delta_y)
        _F_iminus1jminus1 = _F_iplus1jplus1
        _F_iplus1jminus1 = -_F_iplus1jplus1
        _F_iminus1jplus1 = _F_iplus1jminus1

        _equ_RHS_calc_frag = (_F_ij * elev + _F_ijminus1 *
                              elev[cell_neighbors[:, 2]] + _F_ijplus1 *
                              elev[cell_neighbors[:, 0]] + _F_iminus1j *
                              elev[cell_neighbors[:, 3]] + _F_iplus1j *
                              elev[cell_neighbors[:, 1]] + _F_iminus1jminus1 *
                              elev[cell_diagonals[:, 2]] + _F_iplus1jplus1 *
                              elev[cell_diagonals[:, 0]] + _F_iplus1jminus1 *
                              elev[cell_diagonals[:, 1]] + _F_iminus1jplus1 *
                              elev[cell_diagonals[:, 3]])

        # NB- all _z_... and _F_... variables are nnodes long, and thus use
        # real IDs (tho calcs will be flawed for Bnodes)

        # RHS of equ 6 (see para [20])
        _func_on_z = (
            self._rock_density/self._sed_density*self._uplift + _kappa*(
                (_z_xx+_z_yy)/(1.-(_z_x*_z_x+_z_y*_z_y)/_S_crit*_S_crit) +
                2.*(_z_x*_z_x*_z_xx+_z_y*_z_y*_z_yy+2.*_z_x*_z_y*_z_xy) /
                (_S_crit*_S_crit*(1.-(_z_x*_z_x+_z_y*_z_y) /
                                  _S_crit*_S_crit)**2.)))

        # Remember, the RHS is getting wiped each loop as part of
        # self._set_variables()
        # _mat_RHS is ninteriornodes long, but were only working on a
        # ncorenodes long subset here
        _mat_RHS[corenodesbyintIDs] += elev[_core_nodes] + _delta_t * (
            _func_on_z[_core_nodes] - _equ_RHS_calc_frag[_core_nodes])
        low_row = np.vstack((_F_iminus1jminus1, _F_iminus1j,
                             _F_iminus1jplus1)) * -_delta_t
        mid_row = np.vstack((-_delta_t * _F_ijminus1, 1. - _delta_t * _F_ij,
                             -_delta_t * _F_ijplus1))
        top_row = np.vstack((_F_iplus1jminus1, _F_iplus1j,
                             _F_iplus1jplus1)) * -_delta_t
        nine_node_map = np.vstack((low_row, mid_row, top_row)).T
        # ^Note shape is (nnodes,9); it's realID indexed
        core_op_mat_row = np.repeat(corenodesbyintIDs, 9)
        core_op_mat_col = operating_matrix_core_int_IDs.astype(int).flatten()
        core_op_mat_data = nine_node_map[_core_nodes, :].flatten()

        # Now the interior corners; BL,BR,TL,TR
        _mat_RHS[corner_interior_IDs] += (elev[_interior_corners] + _delta_t *
                                          (_func_on_z[_interior_corners] -
                                           _equ_RHS_calc_frag[
                                              _interior_corners]))
        corners_op_mat_row = np.repeat(self.corner_interior_IDs, 4)
        corners_op_mat_col = operating_matrix_corner_int_IDs.astype(
            int).flatten()
        corners_op_mat_data = nine_node_map[_interior_corners, :][
            (np.arange(4).reshape((4, 1)), self.corners_masks)].flatten()
        # ^1st index gives (4,9), 2nd reduces to (4,4), then flattened
        for i in range(4):  # loop over each corner, as so few
            # Note that this ONLY ADDS THE VALUES FOR THE TRUE GRID CORNERS.
            # The sides get done in the edge tests, below.
            if corner_flags[i] == 1:
                true_corner = self.antimask_corner_position[i]
                _mat_RHS[corner_interior_IDs[i]] -= _delta_t * np.sum(
                    nine_node_map[_interior_corners[i], :][
                        corners_antimasks[i, true_corner]] * elev[
                            _interior_corners[i] + modulator_mask[
                                corners_antimasks[i, true_corner]]])
            elif corner_flags[i] == 4 or corner_flags[i] == 3:
                # ^inactive boundary cell
                # Actually the easiest case! Equivalent to fixed gradient,
                # but the gradient is zero, so material only goes in the linked
                # cell. And because it's a true corner, that linked cell
                # doesn't appear in the interior matrix at all!
                pass
            elif corner_flags[i] == 2:
                true_corner = self.antimask_corner_position[i]
                _mat_RHS[corner_interior_IDs[i]] -= _delta_t * np.sum(
                    nine_node_map[_interior_corners[i], :][
                        corners_antimasks[
                            i, true_corner]] * self.fixed_gradient_offset_map[
                                _interior_corners[i] + modulator_mask[
                                    corners_antimasks[i, true_corner]]])
            else:
                raise NameError('''Sorry! This module cannot yet handle fixed
                    gradient or looped BCs...''')
            # Todo: handle these BCs properly, once the grid itself works with
            # them.
            # Can follow old routines; see self.set_bc_cell() commented out
            # method below.

        # Now the edges
        _mat_RHS[bottom_interior_IDs] += elev[_bottom_list] + _delta_t * (
            _func_on_z[_bottom_list] - _equ_RHS_calc_frag[_bottom_list])
        _mat_RHS[top_interior_IDs] += elev[_top_list] + _delta_t * (
            _func_on_z[_top_list] - _equ_RHS_calc_frag[_top_list])
        _mat_RHS[left_interior_IDs] += elev[_left_list] + _delta_t * (
            _func_on_z[_left_list] - _equ_RHS_calc_frag[_left_list])
        _mat_RHS[right_interior_IDs] += elev[_right_list] + _delta_t * (
            _func_on_z[_right_list] - _equ_RHS_calc_frag[_right_list])
        bottom_op_mat_row = np.repeat(bottom_interior_IDs, 6)
        top_op_mat_row = np.repeat(top_interior_IDs, 6)
        left_op_mat_row = np.repeat(left_interior_IDs, 6)
        right_op_mat_row = np.repeat(right_interior_IDs, 6)
        bottom_op_mat_col = self.operating_matrix_bottom_int_IDs.astype(
            int).flatten()
        top_op_mat_col = self.operating_matrix_top_int_IDs.astype(
            int).flatten()
        left_op_mat_col = self.operating_matrix_left_int_IDs.astype(
            int).flatten()
        right_op_mat_col = self.operating_matrix_right_int_IDs.astype(
            int).flatten()
        bottom_op_mat_data = nine_node_map[
            _bottom_list, :][:, self.bottom_mask].flatten()
        top_op_mat_data = nine_node_map[
            _top_list, :][:, self.top_mask].flatten()
        left_op_mat_data = nine_node_map[
            _left_list, :][:, self.left_mask].flatten()
        right_op_mat_data = nine_node_map[
            _right_list, :][:, self.right_mask].flatten()

        if self.bottom_flag == 1:
            # goes to RHS only
            _mat_RHS[bottom_interior_IDs] -= _delta_t * np.sum(
                nine_node_map[_bottom_list, :][:, bottom_antimask] * elev[
                    _bottom_list.reshape((len(_bottom_list), 1)) +
                    (modulator_mask[bottom_antimask]).reshape((1, 3))], axis=1)
            # ^note the broadcasting to (nedge,3) in final fancy index
            # ...& the corners
            edges = [(1, 2), (0, 1), (0, 0), (0, 0)]
            for i in [0, 1]:
                edge_list = edges[i]
                _mat_RHS[corner_interior_IDs[i]] -= _delta_t * np.sum(
                    nine_node_map[_interior_corners[i], :][
                        corners_antimasks[i, edge_list]] * elev[
                            _interior_corners[i] + modulator_mask[
                                corners_antimasks[i, edge_list]]])
            # make dummy array objects for the x,y coords in coo creation of
            # _operating_matrix
            bottom_op_mat_row_add = np.empty(0)
            bottom_op_mat_col_add = np.empty(0)
            bottom_op_mat_data_add = np.empty(0)
        elif self.bottom_flag == 4 or self.bottom_flag == 2:
            # ^i.e., fixed zero gradient (4) or more general case...
            bottom_op_mat_row_add = np.empty(
                (bottom_interior_IDs.size * 3 + 6))
            bottom_op_mat_col_add = np.empty(
                (bottom_interior_IDs.size * 3 + 6))
            bottom_op_mat_data_add = np.empty(
                (bottom_interior_IDs.size * 3 + 6))
            # Equivalent to fixed gradient, but the gradient is zero, so
            # material only goes in the linked cell(i.e., each cell in the
            # op_mat edges points back to itself).
            bottom_op_mat_row_add[
                :(bottom_interior_IDs.size * 3)] = np.repeat(
                    bottom_interior_IDs, 3)
            bottom_op_mat_col_add[
                :(bottom_interior_IDs.size * 3)] = self._realIDtointerior(
                    self.operating_matrix_ID_map[
                        self.bottom_interior_IDs, :][
                            :, self.bottom_mask[0:3]]).flatten()
            bottom_op_mat_data_add[
                :(bottom_interior_IDs.size * 3)] = _delta_t * (
                    nine_node_map[_bottom_list, :][
                        :, bottom_antimask]).flatten()
            # ...& the corners
            this_corner_coords = np.array([0, 1])
            # order is bottom 2 lower left, bottom 2 lower right, lower left
            # true corner, lower right true corner.
            bottom_op_mat_row_add[-6:-2] = np.repeat(
                corner_interior_IDs[this_corner_coords], 2)
            bottom_op_mat_col_add[
                -6:-2] = self.operating_matrix_corner_int_IDs[
                    this_corner_coords.reshape(
                        (2, 1)), this_corner_coords].flatten()
            bottom_op_mat_row_add[-2:] = corner_interior_IDs[
                this_corner_coords]
            bottom_op_mat_col_add[-2:] = self.operating_matrix_corner_int_IDs[
                (this_corner_coords[0], this_corner_coords[0]),
                (this_corner_coords[1], this_corner_coords[1])].flatten()
            bottom_op_mat_data_add[-6:-4] = _delta_t * nine_node_map[
                _interior_corners[0], :][
                    corners_antimasks[0, [1, 2]]].flatten()
            bottom_op_mat_data_add[-4:-2] = _delta_t * nine_node_map[
                _interior_corners[1], :][
                    corners_antimasks[1, [0, 1]]].flatten()
            bottom_op_mat_data_add[-2] = _delta_t * nine_node_map[
                _interior_corners[0], :][corners_antimasks[0, 0]]
            bottom_op_mat_data_add[-1] = _delta_t * nine_node_map[
                _interior_corners[1], :][corners_antimasks[1, 2]]
            if self.bottom_flag == 2:
                # Read the offsets from the map we made in the __init__,
                # use them as constant terms, incorporated into RHS
                _mat_RHS[bottom_interior_IDs] -= _delta_t * np.sum(
                    nine_node_map[_bottom_list, :][:, bottom_antimask] *
                    self.fixed_gradient_offset_map[_bottom_list.reshape(
                        (len(_bottom_list), 1)) + (
                            modulator_mask[bottom_antimask]).reshape(
                                (1, 3))], axis=1)
                # ^note the broadcasting to (nedge,3) in final fancy index
                # ...& the corners
                edges = [(1, 2), (0, 1), (0, 0), (0, 0)]
                for i in [0, 1]:
                    edge_list = edges[i]
                    _mat_RHS[corner_interior_IDs[i]] -= _delta_t * np.sum(
                        nine_node_map[_interior_corners[i], :][
                            corners_antimasks[i, edge_list]] *
                        self.fixed_gradient_offset_map[_interior_corners[
                            i] + modulator_mask[corners_antimasks[
                                i, edge_list]]])
        elif self.bottom_flag == 3:
            # This will handle both top and bottom BCs...
            bottom_op_mat_row_add = np.empty(
                (bottom_interior_IDs.size * 3 + 6))
            bottom_op_mat_col_add = np.empty(
                (bottom_interior_IDs.size * 3 + 6))
            bottom_op_mat_data_add = np.empty(
                (bottom_interior_IDs.size * 3 + 6))
            bottom_op_mat_row_add[
                :(bottom_interior_IDs.size * 3)] = np.repeat(
                    bottom_interior_IDs, 3)
            # ^...put the values in the same places in the operating matrix...
            bottom_op_mat_col_add[
                :(bottom_interior_IDs.size * 3)] = self._realIDtointerior(
                    self.operating_matrix_ID_map[self.top_interior_IDs, :][
                        :, self.top_mask[3:6]]).flatten()
            bottom_op_mat_data_add[:(bottom_interior_IDs.size * 3)] = (
                _delta_t * (nine_node_map[_bottom_list, :][
                    :, bottom_antimask]).flatten())
            # ^...but the values refer to the TOP of the grid
            top_op_mat_row_add = np.empty((top_interior_IDs.size * 3 + 6))
            top_op_mat_col_add = np.empty((top_interior_IDs.size * 3 + 6))
            top_op_mat_data_add = np.empty((top_interior_IDs.size * 3 + 6))
            top_op_mat_row_add[:(top_interior_IDs.size * 3)] = np.repeat(
                top_interior_IDs, 3)
            top_op_mat_col_add[:(
                top_interior_IDs.size * 3)] = self._realIDtointerior(
                    self.operating_matrix_ID_map[self.bottom_interior_IDs, :][
                        :, self.bottom_mask[0:3]]).flatten()
            top_op_mat_data_add[:(top_interior_IDs.size * 3)] = _delta_t * (
                nine_node_map[_top_list, :][:, top_antimask]).flatten()
            # & the corners
            bottom_corner_coords = np.array([0, 1])
            top_corner_coords = np.array([2, 3])
            bottom_op_mat_row_add[-6:-2] = np.repeat(corner_interior_IDs[
                bottom_corner_coords], 2)
            bottom_op_mat_col_add[
                -6:-2] = self.operating_matrix_corner_int_IDs[
                    top_corner_coords.reshape((2, 1)),
                    top_corner_coords].flatten()
            bottom_op_mat_row_add[-2:] = corner_interior_IDs[
                bottom_corner_coords]
            bottom_op_mat_col_add[-2:] = self.operating_matrix_corner_int_IDs[
                (top_corner_coords[0], top_corner_coords[0]), (
                    top_corner_coords[1], top_corner_coords[1])].flatten()
            bottom_op_mat_data_add[-6:-4] = _delta_t * nine_node_map[
                _interior_corners[0], :][corners_antimasks[
                    0, [1, 2]]].flatten()
            bottom_op_mat_data_add[-4:-2] = _delta_t * nine_node_map[
                _interior_corners[1], :][corners_antimasks[
                    1, [0, 1]]].flatten()
            bottom_op_mat_data_add[-2] = _delta_t * nine_node_map[
                _interior_corners[0], :][corners_antimasks[0, 0]]
            bottom_op_mat_data_add[-1] = _delta_t * nine_node_map[
                _interior_corners[1], :][corners_antimasks[1, 2]]
            top_op_mat_row_add[-6:-2] = np.repeat(corner_interior_IDs[
                top_corner_coords], 2)
            top_op_mat_col_add[-6:-2] = self.operating_matrix_corner_int_IDs[
                bottom_corner_coords.reshape((2, 1)),
                bottom_corner_coords].flatten()
            top_op_mat_row_add[-2:] = corner_interior_IDs[top_corner_coords]
            top_op_mat_col_add[-2:] = self.operating_matrix_corner_int_IDs[
                (bottom_corner_coords[0], bottom_corner_coords[0]),
                (bottom_corner_coords[1], bottom_corner_coords[1])].flatten()
            top_op_mat_data_add[-6:-4] = _delta_t * nine_node_map[
                _interior_corners[2], :][corners_antimasks[
                    2, [3, 4]]].flatten()
            top_op_mat_data_add[-4:-2] = _delta_t * nine_node_map[
                _interior_corners[3], :][corners_antimasks[
                    3, [2, 3]]].flatten()
            top_op_mat_data_add[-2] = _delta_t * nine_node_map[
                _interior_corners[2], :][corners_antimasks[2, 2]]
            top_op_mat_data_add[-1] = _delta_t * nine_node_map[
                _interior_corners[3], :][corners_antimasks[3, 4]]
        else:
            raise NameError('''Something is very wrong with your boundary
                            conditions...!''')

        if self.top_flag == 1:
            # goes to RHS only
            _mat_RHS[top_interior_IDs] -= _delta_t * np.sum(nine_node_map[
                _top_list, :][:, top_antimask] * elev[_top_list.reshape(
                    (len(_top_list), 1)) + (modulator_mask[
                        top_antimask]).reshape((1, 3))], axis=1)
            # ...& the corners
            edges = [(0, 0), (0, 0), (3, 4), (2, 3)]
            for i in [2, 3]:
                edge_list = edges[i]
                _mat_RHS[corner_interior_IDs[i]] -= _delta_t * np.sum(
                    nine_node_map[_interior_corners[i], :][corners_antimasks[
                        i, edge_list]] * elev[_interior_corners[
                            i] + modulator_mask[corners_antimasks[
                                i, edge_list]]])
            top_op_mat_row_add = np.empty(0)
            top_op_mat_col_add = np.empty(0)
            top_op_mat_data_add = np.empty(0)
        elif self.top_flag == 4 or self.top_flag == 2:
            top_op_mat_row_add = np.empty((top_interior_IDs.size * 3 + 6))
            top_op_mat_col_add = np.empty((top_interior_IDs.size * 3 + 6))
            top_op_mat_data_add = np.empty((top_interior_IDs.size * 3 + 6))
            # Equivalent to fixed gradient, but the gradient is zero, so
            # material only goes in the linked cell(i.e., each cell in the
            # op_mat edges points back to itself).
            top_op_mat_row_add[:(top_interior_IDs.size * 3)] = np.repeat(
                top_interior_IDs, 3)
            top_op_mat_col_add[
                :(top_interior_IDs.size * 3)] = self._realIDtointerior(
                    self.operating_matrix_ID_map[self.top_interior_IDs, :][
                        :, self.top_mask[3:6]]).flatten()
            top_op_mat_data_add[:(top_interior_IDs.size * 3)] = _delta_t * (
                nine_node_map[_top_list, :][:, top_antimask]).flatten()
            # ...& the corners
            this_corner_coords = np.array([2, 3])
            top_op_mat_row_add[-6:-2] = np.repeat(
                corner_interior_IDs[this_corner_coords], 2)
            top_op_mat_col_add[-6:-2] = self.operating_matrix_corner_int_IDs[
                this_corner_coords.reshape(
                    (2, 1)), this_corner_coords].flatten()
            top_op_mat_row_add[-2:] = corner_interior_IDs[this_corner_coords]
            top_op_mat_col_add[-2:] = self.operating_matrix_corner_int_IDs[
                (this_corner_coords[0], this_corner_coords[0]),
                (this_corner_coords[1], this_corner_coords[1])].flatten()
            top_op_mat_data_add[-6:-4] = _delta_t * nine_node_map[
                _interior_corners[2], :][corners_antimasks[
                    2, [3, 4]]].flatten()
            top_op_mat_data_add[-4:-2] = _delta_t * nine_node_map[
                _interior_corners[3], :][corners_antimasks[
                    3, [2, 3]]].flatten()
            top_op_mat_data_add[-2] = _delta_t * nine_node_map[
                _interior_corners[2], :][corners_antimasks[2, 2]]
            top_op_mat_data_add[-1] = _delta_t * nine_node_map[
                _interior_corners[3], :][corners_antimasks[3, 4]]
            if self.top_flag == 2:
                _mat_RHS[top_interior_IDs] -= _delta_t * np.sum(
                    nine_node_map[_top_list, :][:, top_antimask] *
                    self.fixed_gradient_offset_map[_top_list.reshape(
                        (len(_top_list), 1)) + (modulator_mask[
                            top_antimask]).reshape((1, 3))], axis=1)
                # ...& the corners
                edges = [(0, 0), (0, 0), (3, 4), (2, 3)]
                for i in [2, 3]:
                    edge_list = edges[i]
                    _mat_RHS[corner_interior_IDs[i]] -= _delta_t * np.sum(
                        nine_node_map[_interior_corners[i], :][
                            corners_antimasks[i, edge_list]] *
                        self.fixed_gradient_offset_map[_interior_corners[
                            i] + modulator_mask[corners_antimasks[
                                i, edge_list]]])
        elif self.top_flag == 3:
            pass  # dealt with above
        else:
            raise NameError('''Something is very wrong with your boundary
                            conditions...!''')

        if self.left_flag == 1:
            # goes to RHS only
            _mat_RHS[left_interior_IDs] -= _delta_t * np.sum(nine_node_map[
                _left_list, :][:, left_antimask] * elev[_left_list.reshape(
                    (len(_left_list), 1)) + (modulator_mask[
                        left_antimask]).reshape((1, 3))], axis=1)
            # ...& the corners
            edges = [(3, 4), (0, 0), (0, 1), (0, 0)]
            for i in [0, 2]:
                edge_list = edges[i]
                _mat_RHS[corner_interior_IDs[i]] -= _delta_t * np.sum(
                    nine_node_map[_interior_corners[i], :][corners_antimasks[
                        i, edge_list]] * elev[_interior_corners[
                            i] + modulator_mask[corners_antimasks[
                                i, edge_list]]])
            left_op_mat_row_add = np.empty(0)
            left_op_mat_col_add = np.empty(0)
            left_op_mat_data_add = np.empty(0)
        elif self.left_flag == 4 or self.left_flag == 2:
            left_op_mat_row_add = np.empty((left_interior_IDs.size * 3 + 4))
            left_op_mat_col_add = np.empty((left_interior_IDs.size * 3 + 4))
            left_op_mat_data_add = np.empty(
                (left_interior_IDs.size * 3 + 4))
            # Equivalent to fixed gradient, but the gradient is zero, so
            # material only goes in the linked cell(i.e., each cell in the
            # op_mat edges points back to itself).
            left_op_mat_row_add[:(left_interior_IDs.size * 3)] = np.repeat(
                left_interior_IDs, 3)
            left_op_mat_col_add[
                :(left_interior_IDs.size * 3)] = self._realIDtointerior(
                    self.operating_matrix_ID_map[self.left_interior_IDs, :][
                        :, self.left_mask[::2]]).flatten()
            left_op_mat_data_add[:(left_interior_IDs.size * 3)] = _delta_t * (
                nine_node_map[_left_list, :][:, left_antimask]).flatten()
            # ...& the corners
            this_corner_coords = np.array([0, 2])
            left_op_mat_row_add[-4:] = np.repeat(
                corner_interior_IDs[this_corner_coords], 2)
            left_op_mat_col_add[-4:] = self.operating_matrix_corner_int_IDs[
                this_corner_coords.reshape(
                    (2, 1)), this_corner_coords].flatten()
            left_op_mat_data_add[-4:-2] = _delta_t * nine_node_map[
                _interior_corners[0], :][corners_antimasks[
                    0, [3, 4]]].flatten()
            left_op_mat_data_add[-2:] = _delta_t * nine_node_map[
                _interior_corners[2], :][corners_antimasks[
                    2, [0, 1]]].flatten()
            if self.left_flag == 2:
                _mat_RHS[left_interior_IDs] -= _delta_t * np.sum(
                    nine_node_map[_left_list, :][:, left_antimask] *
                    self.fixed_gradient_offset_map[_left_list.reshape(
                        (len(_left_list), 1)) + (modulator_mask[
                            left_antimask]).reshape((1, 3))], axis=1)
                # ...& the corners
                edges = [(3, 4), (0, 0), (0, 1), (0, 0)]
                for i in [0, 2]:
                    edge_list = edges[i]
                    _mat_RHS[corner_interior_IDs[i]] -= _delta_t * np.sum(
                        nine_node_map[_interior_corners[i], :][
                            corners_antimasks[i, edge_list]] *
                        self.fixed_gradient_offset_map[_interior_corners[
                            i] + modulator_mask[corners_antimasks[
                                i, edge_list]]])
        elif self.left_flag == 3:
            left_op_mat_row_add = np.empty((left_interior_IDs.size * 3 + 4))
            left_op_mat_col_add = np.empty((left_interior_IDs.size * 3 + 4))
            left_op_mat_data_add = np.empty(
                (left_interior_IDs.size * 3 + 4))
            left_op_mat_row_add[:(left_interior_IDs.size * 3)] = np.repeat(
                left_interior_IDs, 3)
            left_op_mat_col_add[
                :(left_interior_IDs.size * 3)] = self._realIDtointerior(
                    self.operating_matrix_ID_map[self.right_interior_IDs, :][
                        :, self.right_mask[1::2]]).flatten()
            left_op_mat_data_add[:(left_interior_IDs.size * 3)] = _delta_t * (
                nine_node_map[_left_list, :][:, left_antimask]).flatten()
            right_op_mat_row_add = np.empty(
                (right_interior_IDs.size * 3 + 4))
            right_op_mat_col_add = np.empty(
                (right_interior_IDs.size * 3 + 4))
            right_op_mat_data_add = np.empty(
                (right_interior_IDs.size * 3 + 4))
            right_op_mat_row_add[:(right_interior_IDs.size * 3)] = np.repeat(
                right_interior_IDs, 3)
            right_op_mat_col_add[
                :(right_interior_IDs.size * 3)] = self._realIDtointerior(
                    self.operating_matrix_ID_map[self.left_interior_IDs, :][
                        :, self.left_mask[::2]]).flatten()
            right_op_mat_data_add[:(right_interior_IDs.size * 3)] = (
                _delta_t * (nine_node_map[_right_list, :][
                    :, right_antimask]).flatten())
            # & the corners
            left_corner_coords = np.array([0, 2])
            right_corner_coords = np.array([1, 3])
            left_op_mat_row_add[-4:] = np.repeat(corner_interior_IDs[
                left_corner_coords], 2)
            left_op_mat_col_add[-4:] = self.operating_matrix_corner_int_IDs[
                right_corner_coords.reshape(
                    (2, 1)), right_corner_coords].flatten()
            left_op_mat_data_add[-4:-2] = _delta_t * nine_node_map[
                _interior_corners[0], :][corners_antimasks[
                    0, [3, 4]]].flatten()
            left_op_mat_data_add[-2:] = _delta_t * nine_node_map[
                _interior_corners[2], :][corners_antimasks[
                    2, [0, 1]]].flatten()
            right_op_mat_row_add[-4:] = np.repeat(corner_interior_IDs[
                right_corner_coords], 2)
            right_op_mat_col_add[-4:] = self.operating_matrix_corner_int_IDs[
                left_corner_coords.reshape(
                    (2, 1)), left_corner_coords].flatten()
            right_op_mat_data_add[-4:-2] = _delta_t * nine_node_map[
                _interior_corners[1], :][corners_antimasks[
                    1, [3, 4]]].flatten()
            right_op_mat_data_add[-2:] = _delta_t * nine_node_map[
                _interior_corners[3], :][corners_antimasks[
                    3, [0, 1]]].flatten()
        else:
            raise NameError('''Something is very wrong with your boundary
                            conditions...!''')

        if self.right_flag == 1:
            # goes to RHS only
            _mat_RHS[right_interior_IDs] -= _delta_t * np.sum(nine_node_map[
                _right_list, :][:, right_antimask] * elev[_right_list.reshape(
                    (len(_right_list), 1)) + (modulator_mask[
                        right_antimask]).reshape((1, 3))], axis=1)
            # ...& the corners
            edges = [(0, 0), (3, 4), (0, 0), (0, 1)]
            for i in [1, 3]:
                edge_list = edges[i]
                _mat_RHS[corner_interior_IDs[i]] -= _delta_t * np.sum(
                    nine_node_map[_interior_corners[i], :][corners_antimasks[
                        i, edge_list]] * elev[_interior_corners[
                            i] + modulator_mask[corners_antimasks[
                                i, edge_list]]])
            right_op_mat_row_add = np.empty(0)
            right_op_mat_col_add = np.empty(0)
            right_op_mat_data_add = np.empty(0)
        elif self.right_flag == 4 or self.right_flag == 2:
            right_op_mat_row_add = np.empty(
                (right_interior_IDs.size * 3 + 4))
            right_op_mat_col_add = np.empty(
                (right_interior_IDs.size * 3 + 4))
            right_op_mat_data_add = np.empty(
                (right_interior_IDs.size * 3 + 4))
            # Equivalent to fixed gradient, but the gradient is zero, so
            # material only goes in the linked cell(i.e., each cell in the
            # op_mat edges points back to itself).
            right_op_mat_row_add[:(right_interior_IDs.size * 3)] = np.repeat(
                right_interior_IDs, 3)
            right_op_mat_col_add[
                :(right_interior_IDs.size * 3)] = self._realIDtointerior(
                    self.operating_matrix_ID_map[self.right_interior_IDs, :][
                        :, self.right_mask[1::2]]).flatten()
            right_op_mat_data_add[:(right_interior_IDs.size * 3)] = (
                _delta_t * (nine_node_map[_right_list, :][
                    :, right_antimask]).flatten())
            # ...& the corners
            this_corner_coords = np.array([1, 3])
            right_op_mat_row_add[-4:] = np.repeat(
                corner_interior_IDs[this_corner_coords], 2)
            right_op_mat_col_add[-4:] = self.operating_matrix_corner_int_IDs[
                this_corner_coords.reshape(
                    (2, 1)), this_corner_coords].flatten()
            right_op_mat_data_add[-4:-2] = _delta_t * nine_node_map[
                _interior_corners[1], :][corners_antimasks[
                    1, [3, 4]]].flatten()
            right_op_mat_data_add[-2:] = _delta_t * nine_node_map[
                _interior_corners[3], :][corners_antimasks[
                    3, [0, 1]]].flatten()
            if self.right_flag == 2:
                _mat_RHS[right_interior_IDs] -= _delta_t * np.sum(
                    nine_node_map[_right_list, :][:, right_antimask] *
                    self.fixed_gradient_offset_map[_right_list.reshape(
                        (len(_right_list), 1)) + (modulator_mask[
                            right_antimask]).reshape((1, 3))], axis=1)
                # ...& the corners
                edges = [(0, 0), (3, 4), (0, 0), (0, 1)]
                for i in [1, 3]:
                    edge_list = edges[i]
                    _mat_RHS[corner_interior_IDs[i]] -= _delta_t * np.sum(
                        nine_node_map[_interior_corners[i], :][
                            corners_antimasks[i, edge_list]] *
                        self.fixed_gradient_offset_map[_interior_corners[
                            i] + modulator_mask[corners_antimasks[
                                i, edge_list]]])
        elif self.top_flag == 3:
            pass  # dealt with above
        else:
            raise NameError('''Something is very wrong with your boundary
                            conditions...!''')

        # new approach using COO sparse matrix requires we build the matrix
        # only now...
        self._operating_matrix = sparse.coo_matrix(
            (np.concatenate((core_op_mat_data, corners_op_mat_data,
                             bottom_op_mat_data, top_op_mat_data,
                             left_op_mat_data, right_op_mat_data,
                             bottom_op_mat_data_add, top_op_mat_data_add,
                             left_op_mat_data_add, right_op_mat_data_add)),
             (np.concatenate((core_op_mat_row, corners_op_mat_row,
                              bottom_op_mat_row, top_op_mat_row,
                              left_op_mat_row, right_op_mat_row,
                              bottom_op_mat_row_add, top_op_mat_row_add,
                              left_op_mat_row_add, right_op_mat_row_add)),
              np.concatenate((core_op_mat_col, corners_op_mat_col,
                              bottom_op_mat_col, top_op_mat_col,
                              left_op_mat_col, right_op_mat_col,
                              bottom_op_mat_col_add, top_op_mat_col_add,
                              left_op_mat_col_add, right_op_mat_col_add)))),
            shape=(n_interior_nodes, n_interior_nodes)).tocsr()
        self._mat_RHS = _mat_RHS


# These methods translate ID numbers between arrays of differing sizes
    def _realIDtointerior(self, ID):
        ncols = self.ncols
        interior_ID = (ID // ncols - 1) * (ncols - 2) + (ID % ncols) - 1
        if (np.any(interior_ID < 0) or
                np.any(interior_ID >= self.ninteriornodes)):
            raise NameError("""One of the supplied nodes was outside the
                            interior grid!""")
        else:
            return interior_ID.astype(int)

    def _interiorIDtoreal(self, ID):
        IGW = self.interior_grid_width
        real_ID = (ID // IGW + 1) * self.ncols + (ID % IGW) + 1
        assert np.all(real_ID < self.nnodes)
        return real_ID.astype(int)

    def _realIDtocore(self, ID):
        ncols = self.ncols
        core_ID = (ID // ncols - 2) * (ncols - 4) + (ID % ncols) - 2
        if np.any(core_ID < 0) or np.any(core_ID >= self.ncorenodes):
            raise NameError("""One of the supplied nodes was outside the
                            core grid!""")
        else:
            return core_ID.astype(int)

    def _coreIDtoreal(self, ID):
        CCW = self.core_cell_width
        real_ID = (ID // CCW + 2) * self.ncols + (ID % CCW) + 2
        assert np.all(real_ID < self.nnodes)
        return real_ID.astype(int)

    def _interiorIDtocore(self, ID):
        IGW = self.interior_grid_width
        core_ID = (ID // IGW - 1) * (self.ncols - 4) + (ID % IGW) - 1
        if np.any(core_ID < 0) or np.any(core_ID >= self.ncorenodes):
            raise NameError("""One of the supplied nodes was outside the
                            core grid!""")
        else:
            return core_ID.astype(int)

    def _coreIDtointerior(self, ID):
        CCW = self.core_cell_width
        interior_ID = (ID // CCW + 1) * (self.ncols - 2) + (ID % CCW) + 1
        assert np.all(interior_ID < self.ninteriornodes)
        return interior_ID.astype(int)

    def diffuse(self, grid_in, elapsed_time, num_uplift_implicit_comps=1):
        """
        This is the "old style" run method of the class, superceded by
        :func:`run_one_step`.
        Takes *grid_in*, the model grid, and *elapsed_time*, the
        total model time elapsed so far.

        *grid_in* must contain the field to diffuse, which defaults to
        'topographic__elevation'. This can be overridden with the
        values_to_diffuse property in the input file.

        See the class docstring for a list of the other properties necessary
        in the input file for this component to run.

        Note that the implicit nature of this component requires it to
        incorporate uplift into its execution in order to stay stable.
        If you only have one module that requires this, do not add uplift
        manually in your loop; this method will include uplift automatically.

        If more than one of your components has this requirement, set
        *num_uplift_implicit_comps* to the total number of components that
        do.
        """
        if self.internal_uplifts:
            # this is adhoc to fix for the duration of Germany visit
            self._uplift = self.inputs.read_float('uplift_rate')
            self._delta_t = self.timestep_in
            self._set_variables(self.grid)
            _interior_elevs = linalg.spsolve(self._operating_matrix,
                                             self._mat_RHS)
            self.grid['node'][self.values_to_diffuse][
                self.interior_IDs_as_real] = _interior_elevs
            grid_in = self.grid
        else:
            self._gear_timestep(self.timestep_in, grid_in)
            for i in range(self.internal_repeats):
                grid_in['node'][self.values_to_diffuse][:] = self.grid['node'][
                    self.values_to_diffuse] + self.uplift_per_step
            # Initialize the variables for the step:
                self._set_variables(grid_in)
                # Solve interior of grid:
                _interior_elevs = linalg.spsolve(self._operating_matrix,
                                                 self._mat_RHS)
                # this fn solves Ax=B for x

                # Handle the BC cells; test common cases first for speed
                self.grid['node'][self.values_to_diffuse][
                    self.interior_IDs_as_real] = _interior_elevs

                # if BC==1 or BC==4, don't need to take any action; in both
                # cases the values are unchanged.
                if self.fixed_grad_BCs_present:
                    self.grid['node'][self.values_to_diffuse][
                        grid_in.fixed_gradient_node_properties[
                            'boundary_node_IDs']] = (
                        self.grid['node'][self.values_to_diffuse][
                            self.grid.fixed_gradient_node_properties[
                                'anchor_node_IDs']] +
                        self.grid.fixed_gradient_node_properties[
                            'values_to_add'])
                if self.looped_BCs_present:
                    self.grid['node'][self.values_to_diffuse][
                        self.grid.looped_node_properties[
                            'boundary_node_IDs']] = self.grid['node'][
                                self.values_to_diffuse][
                                    self.grid.looped_node_properties[
                                        'linked_node_IDs']]

        return self.grid

    def run_one_step(self, dt):
        """Run the diffuser for one timestep, dt.

        This is the primary method of the class.

        Parameters
        ----------
        dt : float (time)
            The imposed timestep.
        """
        if self.internal_uplifts:
            self._delta_t = self.timestep_in
            self._set_variables(self.grid)
            _interior_elevs = linalg.spsolve(self._operating_matrix,
                                             self._mat_RHS)
            self.grid['node'][self.values_to_diffuse][
                self.interior_IDs_as_real] = _interior_elevs
        else:
            self._gear_timestep(dt, self.grid)
            for i in range(self.internal_repeats):
                # Initialize the variables for the step:
                self._set_variables(self.grid)
                # Solve interior of grid:
                _interior_elevs = linalg.spsolve(self._operating_matrix,
                                                 self._mat_RHS)
                # this fn solves Ax=B for x

                # Handle the BC cells; test common cases first for speed
                self.grid['node'][self.values_to_diffuse][
                    self.interior_IDs_as_real] = _interior_elevs

        # if BC==1 or BC==4, don't need to take any action; in both
        # cases the values are unchanged.
        if self.fixed_grad_BCs_present:
            self.grid['node'][self.values_to_diffuse][
                self.grid.fixed_gradient_node_properties[
                    'boundary_node_IDs']] = (
                self.grid['node'][self.values_to_diffuse][
                    self.grid.fixed_gradient_node_properties[
                        'anchor_node_IDs']] +
                self.grid.fixed_gradient_node_properties[
                    'values_to_add'])
        if self.looped_BCs_present:
            self.grid['node'][self.values_to_diffuse][
                self.grid.looped_node_properties[
                    'boundary_node_IDs']] = self.grid['node'][
                        self.values_to_diffuse][
                            self.grid.looped_node_properties[
                                'linked_node_IDs']]



#! /usr/bin/env python
"""Calculate gradients on a raster grid.

Gradient calculators for raster grids
++++++++++++++++++++++++++++++++++++++++++++++

.. autosummary::
    :toctree: generated/

    ~landlab.grid.raster_gradients.calc_grad_at_link
    ~landlab.grid.raster_gradients.calc_grad_at_active_link
    ~landlab.grid.raster_gradients.calc_grad_across_cell_faces
    ~landlab.grid.raster_gradients.calc_grad_across_cell_corners
    ~landlab.grid.raster_gradients.alculate_gradient_along_node_links

"""
import numpy as np

from landlab.core.utils import make_optional_arg_into_id_array
from landlab.grid import gradients
from landlab.grid.base import BAD_INDEX_VALUE, CLOSED_BOUNDARY
from landlab.utils.decorators import use_field_name_or_array


@use_field_name_or_array('node')
def calc_grad_at_link(grid, node_values, out=None):
    """Calculate gradients in node_values at links.

    Construction::

        calc_grad_at_link(grid, node_values, out=None)

    Parameters
    ----------
    grid : RasterModelGrid
        A grid.
    node_values : array_like or field name
        Values at nodes.
    out : ndarray, optional
        Buffer to hold result. If `None`, create a new array.

    Returns
    -------
    ndarray
        Gradients of the nodes values for each link.

    Examples
    --------
    >>> from landlab import RasterModelGrid
    >>> grid = RasterModelGrid((3, 3))
    >>> node_values = [0., 0., 0.,
    ...                1., 3., 1.,
    ...                2., 2., 2.]
    >>> grid.calc_grad_at_link(node_values)
    array([ 0.,  0.,  1.,  3.,  1.,  2., -2.,  1., -1.,  1.,  0.,  0.])

    >>> out = np.empty(grid.number_of_links, dtype=float)
    >>> rtn = grid.calc_grad_at_link(node_values, out=out)
    >>> rtn is out
    True
    >>> out
    array([ 0.,  0.,  1.,  3.,  1.,  2., -2.,  1., -1.,  1.,  0.,  0.])

    >>> grid = RasterModelGrid((3, 3), spacing=(1, 2))
    >>> grid.calc_grad_at_link(node_values)
    array([ 0.,  0.,  1.,  3.,  1.,  1., -1.,  1., -1.,  1.,  0.,  0.])
    >>> _ = grid.add_field('node', 'elevation', node_values)
    >>> grid.calc_grad_at_link('elevation')
    array([ 0.,  0.,  1.,  3.,  1.,  1., -1.,  1., -1.,  1.,  0.,  0.])
    """
    grads = gradients.calc_diff_at_link(grid, node_values, out=out)
    grads /= grid.length_of_link

#    n_vertical_links = (grid.shape[0] - 1) * grid.shape[1]
#    diffs[:n_vertical_links] /= grid.dy
#    diffs[n_vertical_links:] /= grid.dx

    return grads


@use_field_name_or_array('node')
def calc_grad_at_active_link(grid, node_values, out=None):
    """Calculate gradients over active links.

    .. deprecated:: 0.1
        Use :func:`calc_grad_across_cell_faces`
                or :func:`calc_grad_across_cell_corners` instead

    Calculates the gradient in quantity s at each active link in the grid.
    This is nearly identical to the method of the same name in ModelGrid,
    except that it uses a constant node spacing for link length to improve
    efficiency.

    Note that a negative gradient corresponds to a lower node in the
    direction of the link.

    Returns
    -------
    ndarray
        Gradients of the nodes values for each link.

    Examples
    --------
    >>> import numpy as np
    >>> from landlab import RasterModelGrid
    >>> grid = RasterModelGrid(4, 5, 1.0)
    >>> u = [0., 1., 2., 3., 0.,
    ...      1., 2., 3., 2., 3.,
    ...      0., 1., 2., 1., 2.,
    ...      0., 0., 2., 2., 0.]
    >>> grad = grid.calc_grad_at_active_link(u)
    >>> grad # doctest: +NORMALIZE_WHITESPACE
    array([ 1.,  1., -1.,
            1.,  1., -1.,  1.,
           -1., -1., -1.,
            1.,  1., -1.,  1.,
           -1.,  0.,  1.])

    For greater speed, sending a pre-created numpy array as an argument
    avoids having to create a new one with each call:

    >>> grad = np.empty(grid.number_of_active_links)
    >>> rtn = grid.calc_grad_at_active_link(u, out=grad)
    >>> grad # doctest: +NORMALIZE_WHITESPACE
    array([ 1.,  1., -1.,
            1.,  1., -1.,  1.,
           -1., -1., -1.,
            1.,  1., -1.,  1.,
           -1.,  0.,  1.])
    >>> rtn is grad
    True

    >>> grid = RasterModelGrid((3, 3), spacing=(1, 2))
    >>> node_values = [0., 0., 0.,
    ...                1., 3., 1.,
    ...                2., 2., 2.]
    >>> grid.calc_grad_at_active_link(node_values)
    array([ 3.,  1., -1., -1.])

    This function is *deprecated*. Instead, use ``calc_grad_at_link``.

    >>> grid = RasterModelGrid((3, 3), spacing=(1, 2))
    >>> node_values = [0., 0., 0.,
    ...                1., 3., 1.,
    ...                2., 2., 2.]
    >>> grid.calc_grad_at_link(node_values)[grid.active_links]
    array([ 3.,  1., -1., -1.])
    """
    if out is None:
        out = grid.empty(at='active_link')

    if len(out) != grid.number_of_active_links:
        raise ValueError('output buffer does not match that of the grid.')

    # grads = gradients.calculate_diff_at_active_links(grid, node_values,
    #                                                  out=out)
    grads = gradients.calc_diff_at_link(grid, node_values)
    out[:] = grads[grid.active_links]
    out /= grid.length_of_link[grid.active_links]

    return out


@use_field_name_or_array('node')
def calc_grad_across_cell_faces(grid, node_values, *args, **kwds):
    """Get gradients across the faces of a cell.

    Calculate gradient of the value field provided by *node_values* across
    each of the faces of the cells of a grid. The returned gradients are
    ordered as right, top, left, and bottom.

    Note that the returned gradients are masked to exclude neighbor nodes which
    are closed. Beneath the mask is the value -1.

    Construction::

        calc_grad_across_cell_faces(grid, node_values, [cell_ids],
                                             out=None)

    Parameters
    ----------
    grid : RasterModelGrid
        Source grid.
    node_values : array_like or field name
        Quantity to take the gradient of defined at each node.
    cell_ids : array_like, optional
        If provided, cell ids to measure gradients. Otherwise, find gradients
        for all cells.
    out : array_like, optional
        Alternative output array in which to place the result.  Must
        be of the same shape and buffer length as the expected output.

    Returns
    -------
    (N, 4) Masked ndarray
        Gradients for each face of the cell.

    Examples
    --------
    Create a grid with two cells.

    >>> from landlab import RasterModelGrid
    >>> grid = RasterModelGrid(3, 4)
    >>> x = np.array([0., 0., 0., 0.,
    ...               0., 0., 1., 1.,
    ...               3., 3., 3., 3.])

    A decrease in quantity across a face is a negative gradient.

    >>> grid.calc_grad_across_cell_faces(x) # doctest: +NORMALIZE_WHITESPACE
    masked_array(data =
     [[ 1.  3.  0.  0.]
     [ 0.  2. -1. -1.]],
                 mask =
     False,
           fill_value = 1e+20)

    >>> grid = RasterModelGrid((3, 4), spacing=(2, 1))
    >>> grid.calc_grad_across_cell_faces(x) # doctest: +NORMALIZE_WHITESPACE
    masked_array(data =
     [[ 1.   1.5  0.   0. ]
     [ 0.   1.  -1.  -0.5]],
                  mask =
     False,
           fill_value = 1e+20)
    """
    padded_node_values = np.empty(node_values.size + 1, dtype=float)
    padded_node_values[-1] = BAD_INDEX_VALUE
    padded_node_values[:-1] = node_values
    cell_ids = make_optional_arg_into_id_array(grid.number_of_cells, *args)
    node_ids = grid.node_at_cell[cell_ids]

    neighbors = grid.active_neighbors_at_node(node_ids)
    if BAD_INDEX_VALUE != -1:
        neighbors = np.where(neighbors == BAD_INDEX_VALUE, -1, neighbors)
    values_at_neighbors = padded_node_values[neighbors]
    masked_neighbor_values = np.ma.array(
        values_at_neighbors, mask=neighbors == BAD_INDEX_VALUE)
    values_at_nodes = node_values[node_ids].reshape(len(node_ids), 1)

    out = np.subtract(masked_neighbor_values, values_at_nodes, **kwds)

    out[:, (0, 2)] /= grid.dx
    out[:, (1, 3)] /= grid.dy

    return out


@use_field_name_or_array('node')
def calc_grad_across_cell_corners(grid, node_values, *args, **kwds):
    """Get gradients to diagonally opposite nodes.

    Calculate gradient of the value field provided by *node_values* to
    the values at diagonally opposite nodes. The returned gradients are
    ordered as upper-right, upper-left, lower-left and lower-right.

    Construction::

        calc_grad_across_cell_corners(grid, node_values, [cell_ids],
                                               out=None)

    Parameters
    ----------
    grid : RasterModelGrid
        Source grid.
    node_values : array_like or field name
        Quantity to take the gradient of defined at each node.
    cell_ids : array_like, optional
        If provided, cell ids to measure gradients. Otherwise, find gradients
        for all cells.
    out : array_like, optional
        Alternative output array in which to place the result.  Must
        be of the same shape and buffer length as the expected output.

    Returns
    -------
    (N, 4) ndarray
        Gradients to each diagonal node.

    Examples
    --------
    Create a grid with two cells.

    >>> from landlab import RasterModelGrid
    >>> grid = RasterModelGrid(3, 4)
    >>> x = np.array([1., 0., 0., 1.,
    ...               0., 0., 1., 1.,
    ...               3., 3., 3., 3.])

    A decrease in quantity to a diagonal node is a negative gradient.

    >>> from math import sqrt
    >>> grid.calc_grad_across_cell_corners(x) * sqrt(2.)
    array([[ 3.,  3.,  1.,  0.],
           [ 2.,  2., -1.,  0.]])

    >>> grid = RasterModelGrid((3, 4), spacing=(3, 4))
    >>> grid.calc_grad_across_cell_corners(x)
    array([[ 0.6,  0.6,  0.2,  0. ],
           [ 0.4,  0.4, -0.2,  0. ]])
    """
    cell_ids = make_optional_arg_into_id_array(grid.number_of_cells, *args)
    node_ids = grid.node_at_cell[cell_ids]

    values_at_diagonals = node_values[grid._get_diagonal_list(node_ids)]
    values_at_nodes = node_values[node_ids].reshape(len(node_ids), 1)

    out = np.subtract(values_at_diagonals, values_at_nodes, **kwds)
    np.divide(out, np.sqrt(grid.dy ** 2. + grid.dx ** 2.), out=out)

    return out


@use_field_name_or_array('node')
def calc_grad_along_node_links(grid, node_values, *args, **kwds):
    """Get gradients along links touching a node.

    Calculate gradient of the value field provided by *node_values* across
    each of the faces of the nodes of a grid. The returned gradients are
    ordered as right, top, left, and bottom. All returned values follow our
    standard sign convention, where a link pointing N or E and increasing in
    value is positive, a link pointing S or W and increasing in value is
    negative.

    Note that the returned gradients are masked to exclude neighbor nodes which
    are closed. Beneath the mask is the value -1.

    Construction::

        calc_grad_along_node_links(grid, node_values, [cell_ids],
                                            out=None)

    Parameters
    ----------
    grid : RasterModelGrid
        Source grid.
    node_values : array_like or field name
        Quantity to take the gradient of defined at each node.
    node_ids : array_like, optional
        If provided, node ids to measure gradients. Otherwise, find gradients
        for all nodes.
    out : array_like, optional
        Alternative output array in which to place the result.  Must
        be of the same shape and buffer length as the expected output.

    Returns
    -------
    (N, 4) Masked ndarray
        Gradients for each link of the node. Ordering is E,N,W,S.

    Examples
    --------
    Create a grid with nine nodes.

    >>> from landlab import RasterModelGrid
    >>> grid = RasterModelGrid(3, 3)
    >>> x = np.array([0., 0., 0.,
    ...               0., 1., 2.,
    ...               2., 2., 2.])

    A decrease in quantity across a face is a negative gradient.

    >>> grid.calc_grad_along_node_links(x) # doctest: +NORMALIZE_WHITESPACE
    masked_array(data =
     [[-- -- -- --]
     [-- 1.0 -- --]
     [-- -- -- --]
     [1.0 -- -- --]
     [1.0 1.0 1.0 1.0]
     [-- -- 1.0 --]
     [-- -- -- --]
     [-- -- -- 1.0]
     [-- -- -- --]],
                 mask =
     [[ True  True  True  True]
     [ True False  True  True]
     [ True  True  True  True]
     [False  True  True  True]
     [False False False False]
     [ True  True False  True]
     [ True  True  True  True]
     [ True  True  True False]
     [ True  True  True  True]],
           fill_value = 1e+20)

    >>> grid = RasterModelGrid((3, 3), spacing=(2, 4))
    >>> grid.calc_grad_along_node_links(x) # doctest: +NORMALIZE_WHITESPACE
    masked_array(data =
     [[-- -- -- --]
     [-- 0.5 -- --]
     [-- -- -- --]
     [0.25 -- -- --]
     [0.25 0.5 0.25 0.5]
     [-- -- 0.25 --]
     [-- -- -- --]
     [-- -- -- 0.5]
     [-- -- -- --]],
                 mask =
     [[ True  True  True  True]
     [ True False  True  True]
     [ True  True  True  True]
     [False  True  True  True]
     [False False False False]
     [ True  True False  True]
     [ True  True  True  True]
     [ True  True  True False]
     [ True  True  True  True]],
           fill_value = 1e+20)
    """
    padded_node_values = np.empty(node_values.size + 1, dtype=float)
    padded_node_values[-1] = BAD_INDEX_VALUE
    padded_node_values[:-1] = node_values
    node_ids = make_optional_arg_into_id_array(grid.number_of_nodes, *args)

    neighbors = grid.active_neighbors_at_node(node_ids, bad_index=-1)
    values_at_neighbors = padded_node_values[neighbors]
    masked_neighbor_values = np.ma.array(
        values_at_neighbors, mask=values_at_neighbors == BAD_INDEX_VALUE)
    values_at_nodes = node_values[node_ids].reshape(len(node_ids), 1)

    out = np.ma.empty_like(masked_neighbor_values, dtype=float)
    np.subtract(masked_neighbor_values[:, :2],
                values_at_nodes, out=out[:, :2], **kwds)
    np.subtract(values_at_nodes, masked_neighbor_values[:, 2:],
                out=out[:, 2:], **kwds)

    out[:, (0, 2)] /= grid.dx
    out[:, (1, 3)] /= grid.dy

    return out


def calc_unit_normals_at_patch_subtriangles(grid,
                                            elevs='topographic__elevation'):
    """Calculate unit normals on a patch.

    Calculate the four unit normal vectors <a, b, c> to the four possible
    subtriangles of a four-cornered (raster) patch.

    Parameters
    ----------
    grid : RasterModelGrid
        A grid.
    elevs : str or ndarray, optional
        Field name or array of node values.

    Returns
    -------
    (n_TR, n_TL, n_BL, n_BR) : each a num-patches x length-3 array
        Len-4 tuple of the four unit normal vectors <a, b, c> for the four
        possible subtriangles in the patch. Order is (topright, topleft,
        bottomleft, bottomright).

    Examples
    --------
    >>> import numpy as np
    >>> from landlab import RasterModelGrid
    >>> mg = RasterModelGrid((4, 5))
    >>> z = mg.node_x ** 2
    >>> four_tris = mg.calc_unit_normals_at_patch_subtriangles(z)
    >>> type(four_tris) is tuple
    True
    >>> len(four_tris)
    4
    >>> np.allclose(four_tris[0], four_tris[1])
    True
    >>> np.allclose(four_tris[2], four_tris[3])
    True
    >>> np.allclose(four_tris[0], four_tris[2])
    True
    >>> np.allclose(np.square(four_tris[0]).sum(axis=1), 1.)
    True
    >>> four_tris[0]
    array([[-0.70710678,  0.        ,  0.70710678],
           [-0.9486833 ,  0.        ,  0.31622777],
           [-0.98058068,  0.        ,  0.19611614],
           [-0.98994949,  0.        ,  0.14142136],
           [-0.70710678,  0.        ,  0.70710678],
           [-0.9486833 ,  0.        ,  0.31622777],
           [-0.98058068,  0.        ,  0.19611614],
           [-0.98994949,  0.        ,  0.14142136],
           [-0.70710678,  0.        ,  0.70710678],
           [-0.9486833 ,  0.        ,  0.31622777],
           [-0.98058068,  0.        ,  0.19611614],
           [-0.98994949,  0.        ,  0.14142136]])
    """
    try:
        z = grid.at_node[elevs]
    except TypeError:
        z = elevs
    # conceptualize patches as TWO sets of 3 nodes
    # the corners are PQRS, CC from NE
    diff_xyz_PQ = np.empty((grid.number_of_patches, 3))  # TOP
    # ^this is the vector (xQ-xP, yQ-yP, zQ-yP)
    diff_xyz_PS = np.empty((grid.number_of_patches, 3))  # RIGHT
    # we have RS and QR implicitly in PQ and PS - but store them too
    diff_xyz_RS = np.empty((grid.number_of_patches, 3))  # BOTTOM
    diff_xyz_QR = np.empty((grid.number_of_patches, 3))  # LEFT
    P = grid.nodes_at_patch[:, 0]
    Q = grid.nodes_at_patch[:, 1]
    R = grid.nodes_at_patch[:, 2]
    S = grid.nodes_at_patch[:, 3]
    x_P = grid.node_x[P]
    y_P = grid.node_y[P]
    z_P = z[P]
    x_Q = grid.node_x[Q]
    y_Q = grid.node_y[Q]
    z_Q = z[Q]
    x_R = grid.node_x[R]
    y_R = grid.node_y[R]
    z_R = z[R]
    x_S = grid.node_x[S]
    y_S = grid.node_y[S]
    z_S = z[S]
    diff_xyz_PQ[:, 0] = x_Q - x_P
    diff_xyz_PQ[:, 1] = y_Q - y_P
    diff_xyz_PQ[:, 2] = z_Q - z_P
    diff_xyz_PS[:, 0] = x_S - x_P
    diff_xyz_PS[:, 1] = y_S - y_P
    diff_xyz_PS[:, 2] = z_S - z_P
    diff_xyz_RS[:, 0] = x_S - x_R
    diff_xyz_RS[:, 1] = y_S - y_R
    diff_xyz_RS[:, 2] = z_S - z_R
    diff_xyz_QR[:, 0] = x_R - x_Q
    diff_xyz_QR[:, 1] = y_R - y_Q
    diff_xyz_QR[:, 2] = z_R - z_Q
    # make the other ones
    # cross product is orthogonal to both vectors, and is the normal
    # n = <a, b, c>, where plane is ax + by + cz = d
    nhat_topleft = np.cross(diff_xyz_PQ, diff_xyz_QR)  # <a, b, c>
    nhat_bottomright = np.cross(diff_xyz_PS, diff_xyz_RS)
    nhat_topright = np.cross(diff_xyz_PQ, diff_xyz_PS)
    nhat_bottomleft = np.cross(diff_xyz_QR, diff_xyz_RS)
    nmag_topleft = np.sqrt(np.square(nhat_topleft).sum(axis=1))
    nmag_bottomright = np.sqrt(np.square(nhat_bottomright).sum(axis=1))
    nmag_topright = np.sqrt(np.square(nhat_topright).sum(axis=1))
    nmag_bottomleft = np.sqrt(np.square(nhat_bottomleft).sum(axis=1))
    n_TR = nhat_topright/nmag_topright.reshape(grid.number_of_patches, 1)
    n_TL = nhat_topleft/nmag_topleft.reshape(grid.number_of_patches, 1)
    n_BL = nhat_bottomleft/nmag_bottomleft.reshape(
        grid.number_of_patches, 1)
    n_BR = nhat_bottomright/nmag_bottomright.reshape(
        grid.number_of_patches, 1)

    return (n_TR, n_TL, n_BL, n_BR)


def calc_slope_at_patch(grid, elevs='topographic__elevation',
                        subtriangle_unit_normals=None):
    """Calculate the slope (positive magnitude of gradient) at raster patches.

    Returns the mean of the slopes of the four possible patch subtriangles.

    Parameters
    ----------
    grid : RasterModelGrid
        A grid.
    elevs : str or ndarray, optional
        Field name or array of node values.
    subtriangle_unit_normals : tuple of 4 (npatches, 3) arrays (optional)
        The unit normal vectors for the four subtriangles of each patch,
        if already known. Order is TR, TL, BL, BR.

    Returns
    -------
    slopes_at_patch : n_patches-long array
        The slope (positive gradient magnitude) of each patch, in radians.

    Examples
    --------
    >>> import numpy as np
    >>> from landlab import RasterModelGrid
    >>> mg = RasterModelGrid((4, 5))
    >>> z = mg.node_x
    >>> S = mg.calc_slope_at_patch(elevs=z)
    >>> S.size == mg.number_of_patches
    True
    >>> np.allclose(S, np.pi/4.)
    True
    >>> z = mg.node_y**2
    >>> mg.calc_slope_at_patch(elevs=z).reshape((3, 4))
    array([[ 0.78539816,  0.78539816,  0.78539816,  0.78539816],
           [ 1.24904577,  1.24904577,  1.24904577,  1.24904577],
           [ 1.37340077,  1.37340077,  1.37340077,  1.37340077]])
    """
    if subtriangle_unit_normals is not None:
        assert len(subtriangle_unit_normals) == 4
        assert subtriangle_unit_normals[0].shape[1] == 3
        assert subtriangle_unit_normals[1].shape[1] == 3
        assert subtriangle_unit_normals[2].shape[1] == 3
        assert subtriangle_unit_normals[3].shape[1] == 3
        n_TR, n_TL, n_BL, n_BR = subtriangle_unit_normals
    else:
        n_TR, n_TL, n_BL, n_BR = (
            grid.calc_unit_normals_at_patch_subtriangles(elevs))
    dotprod_TL = n_TL[:, 2]  # by definition
    dotprod_BR = n_BR[:, 2]
    dotprod_TR = n_TR[:, 2]
    dotprod_BL = n_BL[:, 2]
    slopes_at_patch_TL = np.arccos(dotprod_TL)
    slopes_at_patch_BR = np.arccos(dotprod_BR)
    slopes_at_patch_TR = np.arccos(dotprod_TR)
    slopes_at_patch_BL = np.arccos(dotprod_BL)
    mean_slope_at_patch = (slopes_at_patch_TR + slopes_at_patch_TL +
                           slopes_at_patch_BL + slopes_at_patch_BR) / 4.

    return mean_slope_at_patch


def calc_grad_at_patch(grid, elevs='topographic__elevation',
                       subtriangle_unit_normals=None,
                       slope_magnitude=None):
    """Calculate the components of the gradient of each raster patch.

    Returns the mean gradient of the four possible patch subtriangles,
    in radians.

    Parameters
    ----------
    grid : RasterModelGrid
        A grid.
    elevs : str or ndarray, optional
        Field name or array of node values.
    subtriangle_unit_normals : tuple of 4 (npatches, 3) arrays (optional)
        The unit normal vectors for the four subtriangles of each patch,
        if already known. Order is TR, TL, BL, BR.
    slope_magnitude : array with size num_patches (optional)
        The mean slope of each patch, if already known. Units must be the
        same as provided here!

    Returns
    -------
    gradient_tuple : (x_component_at_patch, y_component_at_patch)
        Len-2 tuple of arrays giving components of gradient in the x and y
        directions, in the units of *radians*.

    Examples
    --------
    >>> import numpy as np
    >>> from landlab import RasterModelGrid
    >>> mg = RasterModelGrid((4, 5))
    >>> z = mg.node_y
    >>> (x_grad, y_grad) = mg.calc_grad_at_patch(elevs=z)
    >>> np.allclose(y_grad, np.pi/4.)
    True
    >>> np.allclose(x_grad, 0.)
    True
    """
    if subtriangle_unit_normals is not None:
        assert len(subtriangle_unit_normals) == 4
        assert subtriangle_unit_normals[0].shape[1] == 3
        assert subtriangle_unit_normals[1].shape[1] == 3
        assert subtriangle_unit_normals[2].shape[1] == 3
        assert subtriangle_unit_normals[3].shape[1] == 3
        n_TR, n_TL, n_BL, n_BR = subtriangle_unit_normals
    else:
        n_TR, n_TL, n_BL, n_BR = \
            grid.calc_unit_normals_at_patch_subtriangles(elevs)
    if slope_magnitude is not None:
        assert slope_magnitude.size == grid.number_of_patches
        slopes_at_patch = slope_magnitude
    else:
        slopes_at_patch = grid.calc_slope_at_patch(
            elevs=elevs, subtriangle_unit_normals=(n_TR, n_TL, n_BL, n_BR))

    n_sum_x = n_TR[:, 0] + n_TL[:, 0] + n_BL[:, 0] + n_BR[:, 0]
    n_sum_y = n_TR[:, 1] + n_TL[:, 1] + n_BL[:, 1] + n_BR[:, 1]
    theta_sum = np.arctan2(-n_sum_y, -n_sum_x)
    x_slope_patches = np.cos(theta_sum)*slopes_at_patch
    y_slope_patches = np.sin(theta_sum)*slopes_at_patch

    return (x_slope_patches, y_slope_patches)


def calc_slope_at_node(grid, elevs='topographic__elevation',
                       method='patch_mean', return_components=False):
    """Array of slopes at nodes, averaged over neighboring patches.

    Produces a value for node slope (i.e., mean gradient magnitude)
    at each node in a manner analogous to a GIS-style slope map.
    If method=='patch_mean', it averages the gradient on each of the
    patches surrounding the node; if method=='Horn', it returns the
    resolved slope direction. Directional information can still be
    returned through use of the return_components keyword.
    All values are returned in radians, including the components;
    take the tan to recover the rise/run.

    Note that under these definitions, it is not always true that::

        mag, cmp = mg.calc_slope_at_node(z)
        mag**2 == cmp[0]**2 + cmp[1]**2  # only if method=='Horn'

    This is a verion of this code specialized for a raster. It subdivides
    the four square patches around each node into subtriangles,
    in order to ensure more correct solutions that incorporate equally
    weighted information from all surrounding nodes on rough surfaces.

    Parameters
    ----------
    elevs : str or ndarray, optional
        Field name or array of node values.
    method : {'patch_mean', 'Horn'}
        Controls the slope algorithm. Current options are 'patch_mean',
        which takes the mean slope of each pf the four neighboring
        square patches, and 'Horn', which is the standard ArcGIS slope
        algorithm. These produce very similar solutions; the Horn method
        gives a vector mean and the patch_mean gives a scalar mean.
    return_components : bool
        If True, return a tuple, (array_of_magnitude,
        (array_of_slope_x_radians, array_of_slope_y_radians)).
        If false, return an array of floats of the slope magnitude.

    Returns
    -------
    float array or length-2 tuple of float arrays
        If return_components, returns (array_of_magnitude,
        (array_of_slope_x_radians, array_of_slope_y_radians)).
        If not return_components, returns an array of slope magnitudes.

    Examples
    --------
    >>> import numpy as np
    >>> from landlab import RadialModelGrid, RasterModelGrid
    >>> mg = RasterModelGrid((5, 5), 1.)
    >>> z = mg.node_x
    >>> slopes = mg.calc_slope_at_node(elevs=z)
    >>> np.allclose(slopes, np.pi / 4.)
    True
    >>> mg = RasterModelGrid((4, 5), 2.)
    >>> z = - mg.node_y
    >>> slope_mag, cmp = mg.calc_slope_at_node(elevs=z,
    ...                                        return_components=True)
    >>> np.allclose(slope_mag, np.pi / 4.)
    True
    >>> np.allclose(cmp[0], 0.)
    True
    >>> np.allclose(cmp[1], - np.pi / 4.)
    True
    >>> mg = RasterModelGrid((4, 4))
    >>> z = mg.node_x ** 2 + mg.node_y ** 2
    >>> slopes, cmp = mg.calc_slope_at_node(z, return_components=True)
    >>> slopes
    array([ 0.95531662,  1.10991779,  1.32082849,  1.37713803,  1.10991779,
            1.20591837,  1.3454815 ,  1.38904403,  1.32082849,  1.3454815 ,
            1.39288142,  1.41562833,  1.37713803,  1.38904403,  1.41562833,
            1.43030663])
    >>> np.allclose(cmp[0].reshape((4, 4))[:, 0],
    ...             cmp[1].reshape((4, 4))[0, :])  # test radial symmetry
    True
    """
    if method not in ('patch_mean', 'Horn'):
        raise ValueError('method name not understood')
    try:
        patches_at_node = grid.patches_at_node()
    except TypeError:  # was a property, not a fn (=> new style)
        patches_at_node = np.ma.masked_where(
            grid.patches_at_node == -1, grid.patches_at_node,
            copy=False)
    # now, we also want to mask any "closed" patches (any node closed)
    closed_patches = (grid.status_at_node[grid.nodes_at_patch] ==
                      CLOSED_BOUNDARY).sum(axis=1) > 0
    closed_patch_mask = np.logical_or(
        patches_at_node.mask, closed_patches[patches_at_node.data])

    if method == 'patch_mean':
        n_TR, n_TL, n_BL, n_BR = \
            grid.calc_unit_normals_at_patch_subtriangles(elevs)

        mean_slope_at_patches = grid.calc_slope_at_patch(
            elevs=elevs, subtriangle_unit_normals=(n_TR, n_TL, n_BL, n_BR))

        # now CAREFUL - patches_at_node is MASKED
        slopes_at_node_unmasked = mean_slope_at_patches[patches_at_node]
        slopes_at_node_masked = np.ma.array(slopes_at_node_unmasked,
                                            mask=closed_patch_mask)
        slope_mag = np.mean(slopes_at_node_masked, axis=1).data
        if return_components:
            (x_slope_patches, y_slope_patches) = grid.calc_grad_at_patch(
                elevs=elevs, subtriangle_unit_normals=(
                    n_TR, n_TL, n_BL, n_BR),
                slope_magnitude=mean_slope_at_patches)
            x_slope_unmasked = x_slope_patches[patches_at_node]
            x_slope_masked = np.ma.array(x_slope_unmasked,
                                         mask=closed_patch_mask)
            x_slope = np.mean(x_slope_masked, axis=1).data
            y_slope_unmasked = y_slope_patches[patches_at_node]
            y_slope_masked = np.ma.array(y_slope_unmasked,
                                         mask=closed_patch_mask)
            y_slope = np.mean(y_slope_masked, axis=1).data
            mean_grad_x = x_slope
            mean_grad_y = y_slope
    elif method == 'Horn':
        z = np.empty(grid.number_of_nodes + 1, dtype=float)
        mean_grad_x = grid.empty(at='node', dtype=float)
        mean_grad_y = grid.empty(at='node', dtype=float)
        z[-1] = 0.
        try:
            z[:-1] = grid.at_node[elevs]
        except TypeError:
            z[:-1] = elevs
        # proof code for bad indexing:
        diags = grid.diagonal_neighbors_at_node.copy()  # LL order
        orthos = grid.neighbors_at_node.copy()
        # these have closed node neighbors...
        for dirs in (diags, orthos):
            dirs[dirs == BAD_INDEX_VALUE] = -1  # indexing to work
        # now make an array like patches_at_node to store the interim calcs
        patch_slopes_x = np.ma.zeros(patches_at_node.shape, dtype=float)
        patch_slopes_y = np.ma.zeros(patches_at_node.shape, dtype=float)
        diff_E = z[orthos[:, 0]] - z[:-1]
        diff_W = z[:-1] - z[orthos[:, 2]]
        diff_N = z[orthos[:, 1]] - z[:-1]
        diff_S = z[:-1] - z[orthos[:, 3]]
        patch_slopes_x[:, 0] = z[diags[:, 0]] - z[orthos[:, 1]] + diff_E
        patch_slopes_x[:, 1] = z[orthos[:, 1]] - z[diags[:, 1]] + diff_W
        patch_slopes_x[:, 2] = z[orthos[:, 3]] - z[diags[:, 2]] + diff_W
        patch_slopes_x[:, 3] = z[diags[:, 3]] - z[orthos[:, 3]] + diff_E
        patch_slopes_y[:, 0] = z[diags[:, 0]] - z[orthos[:, 0]] + diff_N
        patch_slopes_y[:, 1] = z[diags[:, 1]] - z[orthos[:, 2]] + diff_N
        patch_slopes_y[:, 2] = z[orthos[:, 2]] - z[diags[:, 2]] + diff_S
        patch_slopes_y[:, 3] = z[orthos[:, 0]] - z[diags[:, 3]] + diff_S
        patch_slopes_x /= (2. * grid.dx)
        patch_slopes_y /= (2. * grid.dy)
        patch_slopes_x.mask = closed_patch_mask
        patch_slopes_y.mask = closed_patch_mask
        mean_grad_x = patch_slopes_x.mean(axis=1).data
        mean_grad_y = patch_slopes_y.mean(axis=1).data
        slope_mag = np.arctan(np.sqrt(np.square(mean_grad_x) +
                                      np.square(mean_grad_y)))
        if return_components:
            mean_grad_x = np.arctan(mean_grad_x)
            mean_grad_y = np.arctan(mean_grad_y)

    if return_components:
        return slope_mag, (mean_grad_x, mean_grad_y)

    else:
        return slope_mag



import numpy as np
cimport numpy as np
cimport cython


DTYPE_FLOAT = np.double
ctypedef np.double_t DTYPE_FLOAT_t

DTYPE_INT = np.int
#ctypedef np.longlong_t DTYPE_INT_t
ctypedef np.int_t DTYPE_INT_t


@cython.boundscheck(False)
def adjust_flow_receivers(np.ndarray[DTYPE_INT_t, ndim=1] src_nodes,
                          np.ndarray[DTYPE_INT_t, ndim=1] dst_nodes,
                          np.ndarray[DTYPE_FLOAT_t, ndim=1] z,
                          np.ndarray[DTYPE_FLOAT_t, ndim=1] link_slope,
                          np.ndarray[DTYPE_INT_t, ndim=1] active_links,
                          np.ndarray[DTYPE_INT_t, ndim=1] receiver,
                          np.ndarray[DTYPE_INT_t, ndim=1] receiver_link,
                          np.ndarray[DTYPE_FLOAT_t, ndim=1] steepest_slope):
    """Adjust flow receivers based on link slopes and steepest gradients.

    Parameters
    ----------
    src_nodes : array_like
        Ordered upstream node ids.
    dst_nodes : array_like
        Node ids of nodes receiving flow.
    z : array_like
        Node elevations.
    link_slope : array_like
        Link gradients.
    active_links : array_like
        Link IDs for active links.
    receiver : array_like
        Flow-receiver link IDs.
    steepest_slope : array_like
        Gradient of steepest descent from nodes.
    """
    cdef unsigned int n_nodes = src_nodes.size
    cdef int src_id
    cdef int dst_id

    for i in range(n_nodes):
        src_id = src_nodes[i]
        dst_id = dst_nodes[i]

        if z[src_id] > z[dst_id] and link_slope[i] > steepest_slope[src_id]:
            receiver[src_id] = dst_id
            steepest_slope[src_id] = link_slope[i]
            receiver_link[src_id] = active_links[i]
        elif z[dst_id] > z[src_id] and -link_slope[i] > steepest_slope[dst_id]:
            receiver[dst_id] = src_id
            steepest_slope[dst_id] = - link_slope[i]
            receiver_link[dst_id] = active_links[i]



import numpy as np
from numpy.testing import assert_array_equal
from nose import with_setup
from nose.tools import assert_equal, raises
try:
    from nose.tools import assert_is
except ImportError:
    from landlab.testing.tools import assert_is

from landlab.grid import raster_funcs as rfuncs


def setup_grid():
    """Set up test grid."""
    from landlab import RasterModelGrid
    globals().update({
        'rmg': RasterModelGrid(4, 5),
        'values_at_nodes':  np.arange(20.),
    })


@with_setup(setup_grid)
def test_corner_with_scalar_cell_id():
    """Test using a scalar arg for cell."""
    node_ids = rfuncs.corner_node_at_cell(rmg, np.array([0]), 0)
    assert_array_equal(node_ids, np.array([[2]]))


@with_setup(setup_grid)
def test_face_with_scalar_cell_id():
    """Test using a scalar arg for cell."""
    node_ids = rfuncs.neighbor_node_at_cell(rmg, np.array([0]), 0)
    assert_array_equal(node_ids, np.array([[1]]))


@with_setup(setup_grid)
def test_corner_with_iterable_cell_id():
    """Test using iterable arg for cell."""
    node_ids = rfuncs.corner_node_at_cell(rmg, np.array([0]), (5, ))
    assert_array_equal(node_ids, np.array([[9]]))


@with_setup(setup_grid)
def test_face_with_iterable_cell_id():
    """Test using iterable arg for cell."""
    node_ids = rfuncs.neighbor_node_at_cell(rmg, np.array([0]), (5, ))
    assert_array_equal(node_ids, np.array([[8]]))


@with_setup(setup_grid)
def test_corner_with_no_cell_id():
    """Test without an arg for cells to mean all cells."""
    node_ids = rfuncs.corner_node_at_cell(rmg, np.array([0]))
    assert_array_equal(node_ids, np.array([[2, 3, 4, 7, 8, 9]]).T)


@with_setup(setup_grid)
def test_face_with_no_cell_id():
    """Test without an arg for cells to mean all cells."""
    node_ids = rfuncs.neighbor_node_at_cell(rmg, np.array([0]))
    assert_array_equal(node_ids, np.array([[1, 2, 3, 6, 7, 8]]).T)


@with_setup(setup_grid)
def test_corner_multiple_corners():
    """Test getting nodes for more than one corner."""
    node_ids = rfuncs.corner_node_at_cell(rmg, np.array([0, 1]), (4, ))
    assert_array_equal(node_ids, np.array([[8, 6]]))


@with_setup(setup_grid)
def test_face_multiple_faces():
    """Test getting nodes for more than one corner."""
    node_ids = rfuncs.neighbor_node_at_cell(rmg, np.array([0, 1]), (4, ))
    assert_array_equal(node_ids, np.array([[7, 11]]))


@with_setup(setup_grid)
def test_corner_multiple_corners_and_cells():
    """Test getting nodes for more than one corner and more than one cell."""
    node_ids = rfuncs.corner_node_at_cell(rmg, np.array([0, 1]), (4, 5))
    assert_array_equal(node_ids, np.array([[8, 6], [9, 7]]))


@with_setup(setup_grid)
def test_face_multiple_corners_and_cells():
    """Test getting nodes for more than one corner and more than one cell."""
    node_ids = rfuncs.neighbor_node_at_cell(rmg, np.array([0, 1]), (4, 5))
    assert_array_equal(node_ids, np.array([[7, 11], [8, 12]]))


@with_setup(setup_grid)
def test_corner_type_tuple():
    """Test using tuple as corner arg."""
    node_ids = rfuncs.corner_node_at_cell(rmg, (0, 1), (4, 5))
    assert_array_equal(node_ids, np.array([[8, 6], [9, 7]]))


@with_setup(setup_grid)
def test_face_type_tuple():
    """Test using tuple as face arg."""
    node_ids = rfuncs.neighbor_node_at_cell(rmg, (0, 1), (4, 5))
    assert_array_equal(node_ids, np.array([[7, 11], [8, 12]]))


@with_setup(setup_grid)
def test_corner_type_list():
    """Test using list as corner arg."""
    node_ids = rfuncs.corner_node_at_cell(rmg, [0, 1], (4, 5))
    assert_array_equal(node_ids, np.array([[8, 6], [9, 7]]))


@with_setup(setup_grid)
def test_face_type_list():
    """Test using list as face arg."""
    node_ids = rfuncs.neighbor_node_at_cell(rmg, [0, 1], (4, 5))
    assert_array_equal(node_ids, np.array([[7, 11], [8, 12]]))


@with_setup(setup_grid)
def test_corner_type_scalar():
    """Test using scalar as corner arg."""
    node_ids = rfuncs.corner_node_at_cell(rmg, 2, (4, ))
    assert_array_equal(node_ids, np.array([16]))

    node_ids = rfuncs.corner_node_at_cell(rmg, 2, (4, 5))
    assert_array_equal(node_ids, np.array([16, 17]))

    node_ids = rfuncs.corner_node_at_cell(rmg, 1)
    assert_array_equal(node_ids, np.array([0, 1, 2, 5, 6, 7]))


@with_setup(setup_grid)
def test_face_type_scalar():
    """Test using scalar as face arg."""
    node_ids = rfuncs.neighbor_node_at_cell(rmg, 2, (4, ))
    assert_array_equal(node_ids, np.array([17]))

    node_ids = rfuncs.neighbor_node_at_cell(rmg, 2, (4, 5))
    assert_array_equal(node_ids, np.array([17, 18]))

    node_ids = rfuncs.neighbor_node_at_cell(rmg, 1)
    assert_array_equal(node_ids, np.array([5, 6, 7, 10, 11, 12]))



# -*- coding: utf-8 -*-
"""
Created on Mon Oct 19.

@author: dejh
"""
from __future__ import print_function

import landlab
from landlab import (ModelParameterDictionary, Component, FieldError,
                     FIXED_VALUE_BOUNDARY)

from landlab.utils.decorators import use_file_name_or_kwds, deprecated
from landlab.core.model_parameter_dictionary import MissingKeyError
from landlab.components.flow_routing import (DepressionFinderAndRouter,
                                             FlowRouter)
from landlab.grid.base import BAD_INDEX_VALUE
import numpy as np


class SinkFiller(Component):
    """
    This component identifies depressions in a topographic surface, then fills
    them in in the topography.  No attempt is made to conserve sediment mass.
    User may specify whether the holes should be filled to flat, or with a
    gradient downwards towards the depression outlet. The gradient can be
    spatially variable, and is chosen to not reverse any drainage directions
    at the perimeter of each lake.

    The primary method of this class is 'run_one_step'. 'fill_pits' is a
    synonym.

    Constructor assigns a copy of the grid, and calls the initialize
    method.

    Construction::

        SinkFiller(grid, routing='D8', apply_slope=False, fill_slope=1.e-5):

    Parameters
    ----------
    grid : ModelGrid
        A landlab grid.
    routing : {'D8', 'D4'} (optional)
        If grid is a raster type, controls whether fill connectivity can
        occur on diagonals ('D8', default), or only orthogonally ('D4').
        Has no effect if grid is not a raster.
    apply_slope : bool
        If False (default), leave the top of the filled sink flat. If True,
        apply the slope fill_slope to the top surface to allow subsequent flow
        routing. A test is performed to ensure applying this slope will not
        alter the drainage structure at the edge of the filled region
        (i.e., that we are not accidentally reversing the flow direction
        far from the outlet.)
    fill_slope : float (m/m)
        The slope added to the top surface of filled pits to allow flow
        routing across them, if apply_slope.

    Examples
    --------
    >>> from landlab import RasterModelGrid
    >>> from landlab import BAD_INDEX_VALUE as XX
    >>> from landlab.components import FlowRouter, SinkFiller
    >>> import numpy as np
    >>> lake1 = np.array([34, 35, 36, 44, 45, 46, 54, 55, 56, 65, 74])
    >>> lake2 = np.array([78, 87, 88])
    >>> guard_nodes = np.array([23, 33, 53, 63, 73, 83])
    >>> lake = np.concatenate((lake1, lake2))
    >>> mg = RasterModelGrid((10, 10), 1.)
    >>> z = np.ones(100, dtype=float)
    >>> z += mg.node_x  # add a slope
    >>> z[guard_nodes] += 0.001  # forces the flow out of a particular node
    >>> z[lake] = 0.
    >>> field = mg.add_field('node', 'topographic__elevation', z,
    ...                      units='-', copy=True)
    >>> fr = FlowRouter(mg)
    >>> fr.run_one_step()
    >>> mg.at_node['flow__sink_flag'][mg.core_nodes].sum()
    14
    >>> hf = SinkFiller(mg, apply_slope=False)
    >>> hf.run_one_step()
    >>> np.allclose(mg.at_node['topographic__elevation'][lake1], 4.)
    True
    >>> np.allclose(mg.at_node['topographic__elevation'][lake2], 7.)
    True

    Now reset and demonstrate the adding of an inclined surface:

    >>> field[:] = z
    >>> hf = SinkFiller(mg, apply_slope=True)
    >>> hf.run_one_step()
    >>> hole1 = np.array([4.00007692, 4.00015385, 4.00023077, 4.00053846,
    ...                   4.00038462, 4.00030769, 4.00069231, 4.00061538,
    ...                   4.00046154, 4.00076923, 4.00084615])
    >>> hole2 = np.array([7.4, 7.2, 7.6])
    >>> np.allclose(mg.at_node['topographic__elevation'][lake1], hole1)
    True
    >>> np.allclose(mg.at_node['topographic__elevation'][lake2], hole2)
    True
    >>> fr.run_one_step()
    >>> mg.at_node['flow__sink_flag'][mg.core_nodes].sum()
    0
    """
    _name = 'SinkFiller'

    _input_var_names = ('topographic__elevation',
                        )

    _output_var_names = ('topographic__elevation',
                         'sediment_fill__depth',
                         )

    _var_units = {'topographic__elevation': 'm',
                  'sediment_fill__depth': 'm',
                  }

    _var_mapping = {'topographic__elevation': 'node',
                    'sediment_fill__depth': 'node',
                    }

    _var_doc = {'topographic__elevation': 'Surface topographic elevation',
                'sediment_fill__depth': 'Depth of sediment added at each' +
                                        'node',
                }

    @use_file_name_or_kwds
    def __init__(self, grid, routing='D8', apply_slope=False,
                 fill_slope=1.e-5, **kwds):
        self._grid = grid
        if routing is not 'D8':
            assert routing is 'D4'
        self._routing = routing
        if ((type(self._grid) is landlab.grid.raster.RasterModelGrid) and
                (routing is 'D8')):
            self._D8 = True
            self.num_nbrs = 8
        else:
            self._D8 = False  # useful shorthand for thia test we do a lot
            if type(self._grid) is landlab.grid.raster.RasterModelGrid:
                self.num_nbrs = 4
        self._fill_slope = fill_slope
        self._apply_slope = apply_slope
        self.initialize()

    def initialize(self, input_stream=None):
        """
        The BMI-style initialize method takes an optional input_stream
        parameter, which may be either a ModelParameterDictionary object or
        an input stream from which a ModelParameterDictionary can read values.
        """
        # Create a ModelParameterDictionary for the inputs
        if input_stream is None:
            inputs = None
        elif type(input_stream) == ModelParameterDictionary:
            inputs = input_stream
        else:
            inputs = ModelParameterDictionary(input_stream)

        # Make sure the grid includes elevation data. This means either:
        #  1. The grid has a node field called 'topographic__elevation', or
        #  2. The input file has an item called 'ELEVATION_FIELD_NAME' *and*
        #     a field by this name exists in the grid.
        try:
            self._elev = self._grid.at_node['topographic__elevation']
        except FieldError:
            try:
                self.topo_field_name = inputs.read_string('ELEVATION_' +
                                                          'FIELD_NAME')
            except AttributeError:
                print('Error: Because your grid does not have a node field')
                print('called "topographic__elevation", you need to pass the')
                print('name of a text input file or ModelParameterDictionary,')
                print('and this file or dictionary needs to include the name')
                print('of another field in your grid that contains your')
                print('elevation data.')
                raise AttributeError
            except MissingKeyError:
                print('Error: Because your grid does not have a node field')
                print('called "topographic__elevation", your input file (or')
                print('ModelParameterDictionary) must include an entry with')
                print('the key "ELEVATION_FIELD_NAME", which gives the name')
                print('of a field in your grid that contains your elevation')
                print('data.')
                raise MissingKeyError('ELEVATION_FIELD_NAME')
            try:
                self._elev = self._grid.at_node[self.topo_field_name]
            except AttributeError:
                print('Your grid does not seem to have a node field called',
                      self.topo_field_name)
        else:
            self.topo_field_name = 'topographic__elevation'
        # create the only new output field:
        self.sed_fill_depth = self._grid.add_zeros('node',
                                                   'sediment_fill__depth',
                                                   noclobber=False)

        self._lf = DepressionFinderAndRouter(self._grid, routing=self._routing)
        self._fr = FlowRouter(self._grid, method=self._routing)

    def fill_pits(self, **kwds):
        """
        This is a synonym for the main method :func:`run_one_step`.
        """
        self.run_one_step(**kwds)

    def run_one_step(self, **kwds):
        """
        This is the main method. Call it to fill depressions in a starting
        topography.
        """
        # added for back-compatibility with old formats
        try:
            self._apply_slope = kwds['apply_slope']
        except KeyError:
            pass
        self.original_elev = self._elev.copy()
        # We need this, as we'll have to do ALL this again if we manage
        # to jack the elevs too high in one of the "subsidiary" lakes.
        # We're going to implement the lake_mapper component to do the heavy
        # lifting here, then delete its fields. This means we first need to
        # test if these fields already exist, in which case, we should *not*
        # delete them!
        existing_fields = {}
        spurious_fields = set()
        set_of_outputs = (set(self._lf.output_var_names) |
                          set(self._fr.output_var_names))
        try:
            set_of_outputs.remove(self.topo_field_name)
        except KeyError:
            pass
        for field in set_of_outputs:
            try:
                existing_fields[field] = self._grid.at_node[field].copy()
            except FieldError:  # not there; good!
                spurious_fields.add(field)

        self._fr.route_flow()
        self._lf.map_depressions(pits=self._grid.at_node['flow__sink_flag'],
                                 reroute_flow=True)
        # add the depression depths to get up to flat:
        self._elev += self._grid.at_node['depression__depth']
        # if apply_slope is none, we're now done! But if not...
        if self._apply_slope:
            # new way of doing this - use the upstream structure! Should be
            # both more general and more efficient
            for (outlet_node, lake_code) in zip(self._lf.lake_outlets,
                                                self._lf.lake_codes):
                lake_nodes = np.where(self._lf.lake_map == lake_code)[0]
                lake_perim = self._get_lake_ext_margin(lake_nodes)
                perim_elevs = self._elev[lake_perim]
                out_elev = self._elev[outlet_node]
                lowest_elev_perim = perim_elevs[perim_elevs != out_elev].min()
                # note we exclude the outlet node
                elev_increment = ((lowest_elev_perim-self._elev[outlet_node]) /
                                  (lake_nodes.size + 2.))
                assert elev_increment > 0.
                all_ordering = self._grid.at_node['flow__upstream_node_order']
                upstream_order_bool = np.in1d(all_ordering, lake_nodes,
                                              assume_unique=True)
                lake_upstream_order = all_ordering[upstream_order_bool]
                argsort_lake = np.argsort(lake_upstream_order)
                elevs_to_add = (np.arange(lake_nodes.size, dtype=float) +
                                1.) * elev_increment
                sorted_elevs_to_add = elevs_to_add[argsort_lake]
                self._elev[lake_nodes] += sorted_elevs_to_add
        # now put back any fields that were present initially, and wipe the
        # rest:
        for delete_me in spurious_fields:
            self._grid.delete_field('node', delete_me)
        for update_me in existing_fields.keys():
            self.grid.at_node[update_me][:] = existing_fields[update_me]
        # fill the output field
        self.sed_fill_depth[:] = self._elev - self.original_elev

    @deprecated(use='fill_pits', version=1.0)
    def _fill_pits_old(self, apply_slope=None):
        """

        .. deprecated:: 0.1.38
            Use :func:`fill_pits` instead.

        This is the main method. Call it to fill depressions in a starting
        topography.

        **Output fields**

        *  `topographic__elevation` : the updated elevations
        *  `sediment_fill__depth` : the depth of sediment added at each node

        Parameters
        ----------
        apply_slope : None, bool, or float
            If a float is provided this is the slope of the surface down
            towards the lake outlet. Supply a small positive number, e.g.,
            1.e-5 (or True, to use this default value).
            A test is performed to ensure applying this slope will not alter
            the drainage structure at the edge of the filled region (i.e.,
            that we are not accidentally reversing the flow direction far
            from the outlet.) The component will automatically decrease the
            (supplied or default) gradient a number of times to try to
            accommodate this, but will eventually raise an OverflowError
            if it can't deal with it. If you pass True, the method will use
            the default value of 1.e-5.
        """
        self.original_elev = self._elev.copy()
        # We need this, as we'll have to do ALL this again if we manage
        # to jack the elevs too high in one of the "subsidiary" lakes.
        # We're going to implement the lake_mapper component to do the heavy
        # lifting here, then delete its fields. This means we first need to
        # test if these fields already exist, in which case, we should *not*
        # delete them!
        existing_fields = {}
        spurious_fields = set()
        set_of_outputs = self._lf.output_var_names | self._fr.output_var_names
        try:
            set_of_outputs.remove(self.topo_field_name)
        except KeyError:
            pass
        for field in set_of_outputs:
            try:
                existing_fields[field] = self._grid.at_node[field].copy()
            except FieldError:  # not there; good!
                spurious_fields.add(field)

        self._fr.route_flow()
        self._lf.map_depressions(pits=self._grid.at_node['flow__sink_flag'],
                                 reroute_flow=False)
        # add the depression depths to get up to flat:
        self._elev += self._grid.at_node['depression__depth']
        # if apply_slope is none, we're now done! But if not...
        if apply_slope is True:
            apply_slope = self._fill_slope
        elif type(apply_slope) in (float, int):
            assert apply_slope >= 0.
        if apply_slope:
            # this isn't very efficient, but OK as we're only running this
            # code ONCE in almost all use cases
            sublake = False
            unstable = True
            stability_increment = 0
            self.lake_nodes_treated = np.array([], dtype=int)
            while unstable:
                while 1:
                    for (outlet_node, lake_code) in zip(self._lf.lake_outlets,
                                                        self._lf.lake_codes):
                        self._apply_slope_current_lake(apply_slope,
                                                       outlet_node,
                                                       lake_code, sublake)
                    # Call the mapper again here. Bail out if no core pits are
                    # found.
                    # This is necessary as there are some configs where adding
                    # the slope could create subsidiary pits in the topo
                    self._lf.map_depressions(pits=None, reroute_flow=False)
                    if len(self._lf.lake_outlets) == 0.:
                        break
                    self._elev += self._grid.at_node['depression__depth']
                    sublake = True
                    self.lake_nodes_treated = np.array([], dtype=int)
                # final test that all lakes are not reversing flow dirs
                all_lakes = np.where(self._lf.flood_status <
                                     BAD_INDEX_VALUE)[0]
                unstable = self.drainage_directions_change(all_lakes,
                                                           self.original_elev,
                                                           self._elev)
                if unstable:
                    apply_slope *= 0.1
                    sublake = False
                    self.lake_nodes_treated = np.array([], dtype=int)
                    self._elev[:] = original_elev  # put back init conds
                    stability_increment += 1
                    if stability_increment == 10:
                        raise OverflowError('Filler could not find a stable ' +
                                            'condition with a sloping ' +
                                            'surface!')
        # now put back any fields that were present initially, and wipe the
        # rest:
        for delete_me in spurious_fields:
            self._grid.delete_field('node', delete_me)
        for update_me in existing_fields.keys():
            self.grid.at_node[update_me] = existing_fields[update_me]
        # fill the output field
        self.sed_fill_depth[:] = self._elev - self.original_elev

    def _add_slopes(self, slope, outlet_node, lake_code):
        """
        Assuming you have already run the lake_mapper, adds an incline towards
        the outlet to the nodes in the lake.
        """
        new_elevs = self._elev.copy()
        outlet_coord = (self._grid.node_x[outlet_node],
                        self._grid.node_y[outlet_node])
        lake_nodes = np.where(self._lf.lake_map == lake_code)[0]
        lake_nodes = np.setdiff1d(lake_nodes, self.lake_nodes_treated)
        lake_ext_margin = self._get_lake_ext_margin(lake_nodes)
        d = self._grid.calc_distances_of_nodes_to_point(outlet_coord,
                                                       node_subset=lake_nodes)
        add_vals = slope*d
        new_elevs[lake_nodes] += add_vals
        self.lake_nodes_treated = np.union1d(self.lake_nodes_treated,
                                             lake_nodes)
        return new_elevs, lake_nodes

    def _get_lake_ext_margin(self, lake_nodes):
        """
        Returns the nodes forming the external margin of the lake, honoring
        the *routing* method (D4/D8) if applicable.
        """
        if self._D8 is True:
            all_poss = np.union1d(self._grid.active_neighbors_at_node(lake_nodes),
                                  self._grid._get_diagonal_list(lake_nodes))
        else:
            all_poss = np.unique(self._grid.active_neighbors_at_node(
                lake_nodes))
        lake_ext_edge = np.setdiff1d(all_poss, lake_nodes)
        return lake_ext_edge[lake_ext_edge != BAD_INDEX_VALUE]

    def _get_lake_int_margin(self, lake_nodes, lake_ext_edge):
        """
        Returns the nodes forming the internal margin of the lake, honoring
        the *routing* method (D4/D8) if applicable.
        """
        lee = lake_ext_edge
        if self._D8 is True:
            all_poss_int = np.union1d(self._grid.active_neighbors_at_node(lee),
                                      self._grid._get_diagonal_list(lee))
        else:
            all_poss_int = np.unique(self._grid.active_neighbors_at_node(lee))
        lake_int_edge = np.intersect1d(all_poss_int, lake_nodes)
        return lake_int_edge[lake_int_edge != BAD_INDEX_VALUE]

    def _apply_slope_current_lake(self, apply_slope, outlet_node, lake_code,
                                 sublake):
        """
        Wraps the _add_slopes method to allow handling of conditions where the
        drainage structure would be changed or we're dealing with a sublake.
        """
        while 1:
            starting_elevs = self._elev.copy()
            self._elev[:], lake_nodes = self._add_slopes(apply_slope,
                                                         outlet_node,
                                                         lake_code)
            ext_edge = self._get_lake_ext_margin(lake_nodes)
            if sublake:
                break
            else:
                if not self.drainage_directions_change(lake_nodes,
                                                       starting_elevs,
                                                       self._elev):
                    break
                else:
                    # put the elevs back...
                    self._elev[lake_nodes] = starting_elevs[lake_nodes]
                    # the slope was too big. Reduce it.
                    apply_slope *= 0.1
        # if we get here, either sublake, or drainage dirs are stable

    def drainage_directions_change(self, lake_nodes, old_elevs, new_elevs):
        """
        True if the drainage structure at lake margin changes, False otherwise.
        """
        ext_edge = self._get_lake_ext_margin(lake_nodes)
        if self._D8:
            edge_neighbors = np.hstack((self._grid.active_neighbors_at_node(ext_edge),
                                        self._grid._get_diagonal_list(
                                            ext_edge)))
        else:
            edge_neighbors = self._grid.active_neighbors_at_node(ext_edge).copy()
        edge_neighbors[edge_neighbors == BAD_INDEX_VALUE] = -1
        # ^value irrelevant
        old_neighbor_elevs = old_elevs[edge_neighbors]
        new_neighbor_elevs = new_elevs[edge_neighbors]
        # enforce the "don't change drainage direction" condition:
        edge_elevs = old_elevs[ext_edge].reshape((ext_edge.size, 1))
        cond = np.allclose((edge_elevs >= old_neighbor_elevs),
                           (edge_elevs >= new_neighbor_elevs))
        # if True, we're good, the tilting didn't mess with the fr
        return not cond



#! /usr/bin/env python
import os

import numpy as np
from numpy.testing import assert_array_almost_equal
from nose.tools import assert_true, assert_equal, assert_raises
try:
    from nose.tools import assert_list_equal
except ImportError:
    from landlab.testing.tools import assert_list_equal

from landlab.testing.tools import cdtemp
from landlab.io import read_esri_ascii
from landlab.io.netcdf import read_netcdf
from landlab import RasterModelGrid


def test_save_esri_ascii():
    grid = RasterModelGrid(4, 5, 2.)
    grid.add_field('node', 'air__temperature', np.arange(20.))

    with cdtemp() as _:
        grid.save('test.asc', format='esri-ascii')
        assert_true(os.path.isfile('test.asc'))


def test_add_extension():
    grid = RasterModelGrid(4, 5, 2.)
    grid.add_field('node', 'air__temperature', np.arange(20.))

    with cdtemp() as _:
        grid.save('test', format='esri-ascii')
        assert_true(os.path.isfile('test.asc'))

    with cdtemp() as _:
        grid.save('test', format='netcdf')
        assert_true(os.path.isfile('test.nc'))


def test_replace_extension():
    grid = RasterModelGrid(4, 5, 2.)
    grid.add_field('node', 'air__temperature', np.arange(20.))

    with cdtemp() as _:
        grid.save('test.nc', format='esri-ascii')
        assert_true(os.path.isfile('test.asc'))

    with cdtemp() as _:
        grid.save('test.asc', format='netcdf')
        assert_true(os.path.isfile('test.nc'))


def test_guess_format():
    grid = RasterModelGrid(4, 5, 2.)
    grid.add_field('node', 'air__temperature', np.arange(20.))

    with cdtemp() as _:
        grid.save('test.asc')
        assert_true(os.path.isfile('test.asc'))
        read_esri_ascii('test.asc')

    with cdtemp() as _:
        grid.save('test.nc')
        assert_true(os.path.isfile('test.nc'))
        read_netcdf('test.nc')


def test_names_keyword():
    grid = RasterModelGrid(4, 5, 2.)
    grid.add_field('node', 'air__temperature', np.arange(20.))
    grid.add_field('node', 'land_surface__elevation', np.arange(20.))

    with cdtemp() as _:
        grid.save('test.asc', names='land_surface__elevation')
        assert_true(os.path.isfile('test.asc'))
        read_esri_ascii('test.asc')

    with cdtemp() as _:
        grid.save('test.asc', names=['land_surface__elevation',
                                     'air__temperature'])
        files = ['test_land_surface__elevation.asc',
                 'test_air__temperature.asc']
        for fname in files:
            assert_true(os.path.isfile(fname))
            read_esri_ascii(fname)
